
#define MAX_BUFFER_SIZE 20
#define NUM_SCHEDULERS 2

#define EXCEEDED_BUFFER_LENGTH 1
#define NON_EXISTENT_BUFFER 2
#define FULL_BUFFER 3
#define EMPTY_BUFFER 4
#define SUCCESS 5
#define CANNOT_DISPATCH 6

#define OCCUPIED 10
#define EMPTY 11

#define NO_DEADLINE 0
#define STEP_FUNCTION 20
#define RAMP_FUNCTION 21


/* Comparison scheduling algorithms. */
#define EDF_SP 0
#define FCFS_DD 1

unsigned long current_time; /* Simulated clock time. */

typedef struct task_details
{
	unsigned long arrival_time; /* All tasks READY when they arrive. */
	unsigned long start_time;
	unsigned long run_time;
	unsigned long deadline; 
        /* = 0 (NO_DEADLINE) if non-time-constrained task. */
	char value_func; /* Step or ramp func, if time-constrained. */
	int initial_priority; /* Can be static or dynamic. */
	 	
	int (*func)(); /* Operation specifics - types may change ?? */
	struct robot_args *arg;
} task;

struct queue_info
{
	int front;
	int back;
	int num_queued;
} queue [NUM_SCHEDULERS];

struct buffer_info
{
	int front;
	int back;
	char status; /* Full or empty. */
} buffer [NUM_SCHEDULERS];


extern task *buffer0 [MAX_BUFFER_SIZE]; /* EDF/SP bounded-buffer. */
extern task *buffer1 [MAX_BUFFER_SIZE]; /* FCFS_DD bounded-buffer. */

extern task *schedule_q0 [MAX_BUFFER_SIZE]; /* EDF/SP queue. */
extern task *schedule_q1 [MAX_BUFFER_SIZE]; /* FCFS_DD queue. (_DD = deadline dropping) */

/* Function prototypes. */
void EDF_SP_sched (task *new_task);

void FCFS_DD_sched (task *new_task);

int add_to_buffer (task *new_arrival, int buffer_num, int front,
		 int *back, char *status);

void FCFS_DD_sched (task *new_task);

int dispatch (int queue_num, 
	      task **job);
