/* Asynchronous sound generation. */

#include <unistd.h>
#include <stdio.h>
#include <fcntl.h>
#include <string.h>

#ifdef LINUX
#include <linux/soundcard.h>
#endif

#define SOUNDDIR "./sounds"

#define FALSE       0
#define TRUE        1
#define SBUF_SIZE   32
#define BUFFER_SIZE (1024 * SBUF_SIZE)

int child_pid=0, parent_pid;
static int snd_pipes[2];
static char *Audio_dev = "/dev/audio";
static int  Audio_fd;

/* size should depend on sample_rate. */
static unsigned char   	buf [BUFFER_SIZE];       

int setUpAudioSystem () {
  int err, cnt;

  if (child_pid == 0) {
    if (pipe ((int *)&snd_pipes)) {
      fprintf (stderr, "Cannot create pipes to communicate with child.\n");
      return FALSE;
    }
    if ((child_pid = fork())) {
      /* Successfully forked child. */
      parent_pid = getpid();
      return TRUE;
    }
    else {
      /* Child process. */
      static char string[81]; 
      static char *str = SOUNDDIR;
      static char soundfile[256];
      static int ifd;

      child_pid = getpid();
      
      do {
	read (snd_pipes[0], &string, 80);
	if (!strncmp (string, "EXIT", strlen("EXIT"))) {

#ifdef LINUX	  
	  /* Make sure that the audio device is flushed and reset. */
	  ioctl (Audio_fd, SNDCTL_DSP_RESET, 0);
#endif

	/* Close the sound file. */
	  close (ifd);
	  close (Audio_fd);
	  return 0;
	}	

	/* Try to open the audio device. */
	if (!(Audio_fd = open (Audio_dev, O_WRONLY))) 	
	  continue;
	
	/* Must be a sound file name. */
	sprintf (soundfile, "%s/%s.au", str, string);
      
	/* Open the sound file for reading. */
	if ((ifd = open (soundfile, O_RDONLY, 0)) < 0)
	  fprintf (stderr, "Unable to open sound file %s.\n", soundfile);

	/* At this point, we're all ready to copy the data. */
	while ((cnt = read (ifd, (char *) buf, BUFFER_SIZE)) >= 0) {
	  err = write (Audio_fd, (char *)buf, cnt);
	  
	  if (err != cnt)
	    fprintf (stderr, "Problem while writing to sound device.\n");
	  if (cnt == 0) 
	    break;
	}

	if (cnt < 0)
	  /* Some error - while reading - notify user. */
	  fprintf (stderr, "Problem while reading soundfile %s.\n", 
		   soundfile);
	
#ifdef LINUX
	/* Flush any audio activity. */
	if (ioctl (Audio_fd, SNDCTL_DSP_SYNC, 0) < 0)
	  fprintf (stderr, "Unable to flush audio device.\n");
#endif

	/* Close the sound file. */
	close (ifd);
	close (Audio_fd);
      } while (TRUE);
    }
  }
}


void freeAudioSystem () {
  /* Close the audio device. */
  if (write (snd_pipes[1], "EXIT", strlen("EXIT")) < 0)
    fprintf (stderr, "Cannot write to audio device.\n");
}


void playSoundFile (char *filename) {
  static char soundfile[80];

  strcpy (soundfile, filename);
  if (write (snd_pipes[1], soundfile, 80) < 0)
    fprintf (stderr, "Cannot write to audio device.\n");
}


