#include <pwd.h>
#include <stdlib.h>
#include "Xdata.h"
#include "socklink.h"
#include "consistency.h"

#define FULL_BANDWIDTH    (8*MAX_IMAGE_SIZE)
#define REDUCED_BANDWIDTH MAX_IMAGE_SIZE

extern unsigned long current_time;
extern char *pause_string;
extern char *progname;
extern int sound;
extern int team_size;
extern int stop;
extern int cumulative_tot_score [NUM_TEAMS];
extern int session_tot_score [NUM_TEAMS];
extern char *sector; 
extern int sconn_id;

FILE *debug_file;

static int rconn_id;
static textmsg_t rtext, stext;  /* Receive/send text msg buffers.           */
static datamsg_t rdata, sdata;	/* Receive/send data msg buffers.           */
static char msg_buf [NET_MAX_PACKET_SIZE];
databuf_t databuf [NUM_TEAMS];
static int msg_len;
static int active[NUM_TEAMS];	/* Array of clients still playing game and
				   requiring 'free_token' to access shared 
				   data.                                    */

static int goal_score;
client_t client [NUM_TEAMS];
int num_clients;
int free_token = FALSE;		 /* Controls connection sequence.           */
int free_data_token = TRUE;	 /* Prevents data-races.                    */
int conn_established[NUM_TEAMS]; /* Keeps tally of who has connected to whom*/

extern void update_time ();
extern void refreshSWindow (int *cumulative_tot_score, 
			    int *session_tot_score);
extern void redrawGoal (char *sector, coordinate goal);
extern void getRobotInfo (int x, int y, 
			  int *team_number, 
			  int *robot_number);
void sendToken (int *token, int active_client);
int socketHandle (int timeout);
void resync (int *exchange_list, int resync_flag);
void retrieveDataMsg (int slot_id);
void applyDataMsg (char *buf, datamsg_t *data);

void sendMsg (int conn_id, 
	      void *msg, 
	      int msg_size,
	      char *error_msg) {
  int num_tries = 0;

  while (Net_Send (conn_id, 
		   msg, 
		   msg_size) != 
	 NET_SUCCESS) {
    if (num_tries < 3) {
      fprintf (stderr, 
	       "%s: (%s) Failure to send to host "
	       "with connection id:%d\n"
	       "Trying again...\n",
	       progname,
	       error_msg,
	       conn_id);
      num_tries++;
      /* Wait random amount of time before retrying. */ 
      snooze ((unsigned) (random()%1000000));
    }
    shutdown (1);
  }

}


void signon (int team_size, int port) {
  
  char s[10];
  char str[80];
  int i;

  memset (&stext, -1, sizeof (stext));
  stext.m                = OUTSWAP (text);
  stext.id               = OUTSWAP (id);
  stext.type             = OUTSWAP (SIGNON);
  stext.info.signon.port = OUTSWAP (port);
  gethostname (stext.info.signon.hostname, HOST_LEN);
  strncpy (stext.info.signon.username, 
	   ((struct passwd *) getpwuid (getuid()))->pw_name, NAME_LEN);
  stext.info.signon.team_size = OUTSWAP (team_size);

  sendMsg (sconn_id, &stext, sizeof (stext), "signon");
  
  /*ET( MessageBox );*/

  printf ("Waiting for accept msg\n");
  while (socketHandle(0) != ACCEPT)
    while (eventCheck (FALSE, &pause_string));
  printf ("Waiting for setup to complete\n");
  while (socketHandle(0) != SETUP_COMPLETE);
  
#ifdef DEBUG
  sprintf (s, "debug%d", id);
  debug_file = fopen (s, "w+");
#endif

  switch (con_protocol) {
  case BSYNC:
    strcpy (s, "BSYNC");
    break;
  case MSYNC:
    strcpy (s, "MSYNC");
    break;
  case ENTRY:
    strcpy (s, "entry");
    break;
  default:
    break;
  }
  
  printf ("Signon accepted.\n"
	  "Client ID = %d\n"
	  "Starting goal score = %d\n"
	  "num_clients = %d\n"
	  "Consistency protocol = %s\n",
	  id, goal_score, num_clients, s);
  
  ET( ClearStatusWindow .l );
  for (i = 0; i < num_clients; i++) {
    printf ("client[%d].port = %d\n"
	    "client[%d].registered = %d\n"
	    "client[%d].username = %s\n"
	    "client[%d].hostname = %s\n",
	    i, client[i].port,
	    i, client[i].registered,
	    i, client[i].username,
	    i, client[i].hostname);
    
    sprintf (str, "%s%c%s\t%d\t%d\n", 
	     client[i].username,
	     '@',
	     client[i].hostname,
	     session_tot_score[i], 
	     cumulative_tot_score[i]);
    ET( InitStatusWindow .l "%q(str)" );
  }
    
  /*ET( destroy .mbox );*/
}


/* Send signoff to server and shutdown. */
void signoff (int mode, char *reason) { 

  memset (&stext, -1, sizeof (stext));

  stext.m                   = OUTSWAP (text);
  stext.id                  = OUTSWAP (id);
  stext.type                = OUTSWAP (SIGNOFF);
  stext.info.signoff.flag   = OUTSWAP (mode);
  strcpy (stext.info.signoff.reason, reason);

  sendMsg (sconn_id, &stext, sizeof (stext), "signoff");
  
  stop = TRUE;

  if (mode == S_QUIT)
    shutdown (0);
}


/* Establish connections between all clients in distributed cluster. */
int negotiate (int id) {
  int i;
  int conn_id;
  int num_tries=0;
  static int flag = TRUE;
  
  if (flag) {
    /* Pass 'free_token' to client with ID 0. 'free_token' is 
       needed to access shared memory. CAVEAT: To avoid data-race
       problems, we might have rx'ed 'free_token' from process id = 0
       before we call this function. If "free_token' is TRUE, we did
       receive the token, so we don't want to reset its value here! */
    if (!free_token)
      free_token = (id == 0) ? TRUE : FALSE;
    flag = FALSE; /* Never need to setup connections again. */
  }

  /* Use token to specify ordering for opening connections. 
     If token is not yet free, wait to receive token before establishing
     connections. */
  if (free_token == FALSE)
    if (socketHandle(0) != TOKEN)
      return (FALSE);

  if (conn_established[id] == FALSE) 
    for (i = 0; i < num_clients; i++) {
      if (i == id)
	continue;
      
      while (Net_Open (client[i].hostname, client[i].port, &conn_id) != 
	     NET_SUCCESS) {
	if (num_tries < 3) {
	  fprintf (stderr, 
		   "%s: Failed to open connection to %s on port %d.\n"
		   "Trying to connect again...\n", 
		   progname,
		   client[i].hostname, 
		   client[i].port);
	  num_tries++;
	  /* Wait random amount of time before retrying. 
	     Reduces window for concurrent connections by other hosts. */
	  snooze ((unsigned) (random()%1000000));
	}
	else
	  shutdown (1);
      }
      client[i].conn_id = conn_id;
    }
  /* Pass token to next client in ring to establish connections. */
  conn_established[id] = TRUE;
  sendToken (&free_token, TRUE);
  
  for (i = 0; i < num_clients; i++)
    if (conn_established[i] == FALSE)
      return (FALSE);

  printf ("(negotiate)\n");
  for (i = 0; i < num_clients; i++)
    printf ("client[%d].conn_id = %d\n",
	    i, client[i].conn_id);

  /*memset (&stext, -1, sizeof (stext));
  stext.m                = OUTSWAP (text);
  stext.id               = OUTSWAP (id);
  stext.type             = OUTSWAP (NEGOTIATE);
  sendMsg (sconn_id, &stext, sizeof (stext), "negotiate");

  while (socketHandle(0) != NEGOTIATE);*/
  return (TRUE);
}


void resync (int *exchange_list, int resync_flag) {

  int i;
  int min_dist[NUM_TEAMS];
  int exchange_count[NUM_TEAMS];
  int flag = FALSE;

  /* Keep note of which clients we are sync'ing with.
   i.e. expecting exchanges with. */
  for (i = 0; i < num_clients; i++)
    exchange_count[i] = exchange_list[i];

  exchange_count[id] = FALSE;
  for (i = 0; i < num_clients; i++)
    flag = flag || exchange_count[i];
  if (!flag)
    return;
  
  /* Send SYNC message to remote client to inform that client that
   this process has finished sending its exchange messages. Only
   send SYNC messages to clients that need exchanges at this time. */

  memset (&stext, -1, sizeof (stext));
  stext.m            = OUTSWAP (text);
  stext.id           = OUTSWAP (id);
  stext.type         = OUTSWAP (SYNC);
  stext.info.sync.ts = OUTSWAP (current_time);
  if (resync_flag)
    stext.info.sync.resync_flag = OUTSWAP (TRUE);
  else
    stext.info.sync.resync_flag = OUTSWAP (FALSE);

  for (i = 0; i < num_clients; i++) {
    if ((client[i].registered == FALSE) || 
	(i == id) || (exchange_list[i] == FALSE))
      continue;

    sendMsg (client[i].conn_id, &stext, sizeof (stext), "resync");

#ifdef DEBUG
    fprintf (debug_file, "Sending SYNC msg to client %d\n", i);
#endif
  }

  if (!resync_flag)
    return;

  /* Check for SYNC messages that have already arrived from 
     remote hosts. */
  for (i = 0; i < num_clients; i++)
    if (sync_list[i].sync_flag && exchange_count[i]) {
      exchange_count[i] = FALSE;
      sync_list[i].pending--;
      if (sync_list[i].pending <= 0) {
	sync_list[i].pending = 0;
	sync_list[i].sync_flag = FALSE;
      }

#ifdef DEBUG      
      fprintf (debug_file, 
	       "Rx'ed SYNC msg from client %d \n",
	       i);
#endif
    }
  
  /* Wait for exchange messages from all remote clients to arrive.
     In the meantime process X events and poll the sockets for each
     remote client. */
  while (TRUE) {
    flag = FALSE;
    for (i = 0; i < num_clients; i++)
      flag = flag || exchange_count[i];
    if (!flag)
      break;

    eventCheck (FALSE, &pause_string);
    if (socketHandle(0) == SYNC)
      for (i = 0; i < num_clients; i++)
	if (sync_list[i].sync_flag && exchange_count[i]) {
	  exchange_count[i] = FALSE;
	  sync_list[i].pending--;
	  if (sync_list[i].pending <= 0) {
	    sync_list[i].pending = 0;
	    sync_list[i].sync_flag = FALSE;
	  }

#ifdef DEBUG
	  fprintf (debug_file, 
		   "Rx'ed SYNC msg from client %d \n",
		   i);
#endif
	}
#ifdef DEBUG
    fprintf (debug_file, 
	     "syncing %lu\n", current_time);
#endif
  }
  
#ifdef DEBUG
  fprintf (debug_file,
	   "synced %lu\n", current_time);
#endif

  /* Re-establish the next exchange-times with the clients we 
     just exchanged with. 
     STEP 1: Delete old exchange-times for the clients we just 
     received from. 
     STEP 2: Calculate new exchange-times based on closest 
     distance to remote clients. 
     STEP 3: Add new exchange-times to list. */

  for (i = 0; i < num_clients; i++)
    if (exchange_list[i]) {
      exchange_list[i] = FALSE;
      delExchangeTimeFromList (i, &el_front);
      
      if (resync_flag && sync_list[i].resync_flag && (i != id)) {
	calcShortestDist (i, min_dist);
	
	if (con_protocol == MSYNC) 
	  addExchangeTimeToList (i, 
				 current_time + 
				 floor ((double)min_dist[i]/2),
				 &el_front);
	else if (con_protocol == BSYNC)
	  addExchangeTimeToList (i, 
				 current_time + 1,
				 &el_front);
      }
    }
}


void sendEvent (int (*func)(struct robot_args *), 
		struct robot_args *arg,
		int *exchange_list,
		int resync_flag,
		send_t how) { 

  int i, j;
  int x, y;
  datareq_t client_list[NUM_TEAMS][MAX_RANGE];
  int var_data_size;
  sched_t *exchange_ptr;
  char *buf;

  update_time(); /* Note the updating of time is before
		    we send events to other clients. */

  /* Check to see if we can apply any buffered data msgs now
     that we have updated the local clock.*/
  for (i = 0; i < num_clients; i++)
    if (i != id)
      retrieveDataMsg (i);

  /* Check to see if it's time to exchange event data with any 
     other clients. The global variable 'el_front' points to a linked
     list of (exchange-time, client) pairs. This list is ordered by
     increasing time so we can traverse the list until we reach 
     the first non-NULL member that has a time greater than 
     current_time. */
  
  for (i = 0; i < num_clients; i++) {
    if (how == multicast)
      exchange_list[i] = FALSE;
    else
      exchange_list[i] = TRUE;
  }

  /* Build exchange-list of all clients we need to exchange data with. */
  exchange_ptr = el_front;     /* Don't mess with 'el_front' just yet. */
  while (exchange_ptr != NULL) {
    if (how == multicast)
      if (exchange_ptr->time > current_time)
	break;

    exchange_list[exchange_ptr->client_id] = TRUE;
    exchange_ptr = exchange_ptr->next;
  }  
  if (exchange_list[id]) {
    exchange_list[id] = FALSE;
    delExchangeTimeFromList (id, &el_front);
  }

  /* If we have any buffered events, exchange appropriate events with
   exchange-clients that must be exchanged with now. */

  for (i = 0; i < num_clients; i++)
    for (j = 0; j < team_size; j++) {
      if (event_buffer[i][j] && exchange_list[i]) {	
	/* Send client i details about robot j. */
	event_buffer[i][j] = FALSE;
	if ((client[i].registered == FALSE) || (i == id)) 
	  continue;

	memset (&sdata, -1, sizeof (sdata));
	sdata.m            = OUTSWAP (data);
	sdata.id           = OUTSWAP (id);
	sdata.ts           = OUTSWAP (current_time);
	sdata.rinfo.team_size      = OUTSWAP (team_size);
	sdata.rinfo.robot_number   = OUTSWAP (j);
	sdata.rinfo.robot_life     = OUTSWAP (robot_life[id][j]);
	sdata.rinfo.killed_neg     = OUTSWAP (killed_neg[id][j]);
	sdata.position.x   = 
	  OUTSWAP (robot_position[id][j].x);
	sdata.position.y   = 
	  OUTSWAP (robot_position[id][j].y);
	sdata.position.orientation = OUTSWAP (robot_front[id][j]);
	sdata.position.object_type = 
	  OUTSWAP (ENEMY_ROBOT);
	sdata.size = 0;		/* No variable data to tag to message. */
	
	sendMsg (client[i].conn_id, &sdata, sizeof (sdata), "sendEvent");
      }
    }
  
  /* Right now, we always broadcast fire_missile events. */
  if (func == fire_missile) {

#ifndef TEST
    memset (&stext, -1, sizeof (stext));
    stext.m    = OUTSWAP (text);
    stext.type = OUTSWAP (A_MISSILE);
    stext.id   = OUTSWAP (id);
    stext.info.missile.robot_number = OUTSWAP (arg->robot_num);
    stext.info.missile.active       = OUTSWAP (TRUE);
    stext.info.missile.x            = 
      OUTSWAP (robot_position[id][arg->robot_num].x);
    stext.info.missile.y            = 
      OUTSWAP (robot_position[id][arg->robot_num].y);
    stext.info.missile.orientation  = 
      OUTSWAP (/*-1*/robot_front[id][arg->robot_num]);

    for (i = 0; i < num_clients; i++) {
      if ((client[i].registered == FALSE) || 
	  (i == id))
	continue;

      sendMsg (client[i].conn_id, &stext, sizeof (stext), "sendEvent");
    }
#endif

  }
  else if (func) { /* Event must have been robot walk or rotation. */
    if (func == walk) {
      /* Send other clients info about the block 
	 from where the robot walked. We always send this, but
	 it would be better to let the recipient work out that the
	 old location of the robot is no longer valid. This will be
	 changed in due course to cut down on message transfers. */
      switch (arg->value) {
      case NORTH:
	x = robot_position[id][arg->robot_num].x;
	y = robot_position[id][arg->robot_num].y+1;
	break;
      case EAST:
	x = robot_position[id][arg->robot_num].x-1;
	y = robot_position[id][arg->robot_num].y;
	break;
      case SOUTH:
	x = robot_position[id][arg->robot_num].x;
	y = robot_position[id][arg->robot_num].y-1;
	break;
      case WEST:
	x = robot_position[id][arg->robot_num].x+1;
	y = robot_position[id][arg->robot_num].y;
	break;
      }
      memset (&sdata, -1, sizeof (sdata));
      sdata.m            = OUTSWAP (data);
      sdata.id           = OUTSWAP (id);
      sdata.ts           = OUTSWAP (current_time);
      sdata.position.orientation = OUTSWAP (0); /* Don't-care value. */
      sdata.position.object_type = OUTSWAP (world[x][y]);
      sdata.position.x = OUTSWAP (x);
      sdata.position.y = OUTSWAP (y);
      sdata.size = 0;		/* No variable data to tag to message. */

      for (i = 0; i < num_clients; i++) {
	if ((client[i].registered == FALSE) || (i == id))
	  continue;
	
	sendMsg (client[i].conn_id, &sdata, sizeof (sdata), "sendEvent");
      }
    }

    /* Send info about new position or orientation of robot. */
    memset (&sdata, -1, sizeof (sdata));
    sdata.m            = OUTSWAP (data);
    sdata.id           = OUTSWAP (id);
    sdata.ts           = OUTSWAP (current_time);
    sdata.rinfo.team_size      = OUTSWAP (team_size);
    sdata.rinfo.robot_number   = OUTSWAP (arg->robot_num);
    sdata.rinfo.robot_life     = OUTSWAP (robot_life[id][arg->robot_num]);
    sdata.rinfo.killed_neg     = OUTSWAP (killed_neg[id][arg->robot_num]);
    sdata.position.x   = 
      OUTSWAP (robot_position[id][arg->robot_num].x);
    sdata.position.y   = 
      OUTSWAP (robot_position[id][arg->robot_num].y);
    sdata.position.orientation = OUTSWAP (robot_front[id][arg->robot_num]);

#if 0  /* UNKNOWN images and spatially-proportional b/w disabled. */

    /* Establish list of clients that need to know full detail about images
       sent with event information. */
    
    for (i = 0; i < FULL_ROTATION; i++)
      look (arg->robot_num, sighted_flag, i, 6, client_list);

#endif

    /* NOTE: This disables UNKNOWN images. */
    for (i = 0; i < num_clients; i++)
      client_list[i][0].flag = ENEMY_ROBOT;

    for (i = 0; i < num_clients; i++) {
      if ((client[i].registered == FALSE) || (i == id))
	continue;

      if (exchange_list[i] == FALSE) { 
	/* Buffer event and delay sending. */
	event_buffer[i][arg->robot_num] = TRUE;
	continue;
      }
      
      /* Send appropriate (pixmap) image information, corresponding to event.
	 NOTE: 
	 1. All robots in this client's team look like enemies to other 
	 clients.
	 2. Cannot send XImages, since we cannot get an image from an
	 inactive sector of the battlefield window, FWin. 
	 3. The dispatcher will have manipulated the robot causing the
	 event so that if the robot rotated it will be pointing in 
	 the right direction at this point. */

      if (client_list[i][0].flag == ENEMY_ROBOT || 
	  client_list[i][0].flag == UNKNOWN) {
	sdata.position.object_type = OUTSWAP (ENEMY_ROBOT);

	sdata.size = OUTSWAP (FULL_BANDWIDTH);
	var_data_size = FULL_BANDWIDTH;

	switch (robot_front[id][arg->robot_num]) {
	case NORTH:
	  bcopy (brobotn_xpm, msg_buf+sizeof (sdata), FULL_BANDWIDTH);
	  break;
	case EAST:
	  bcopy (brobote_xpm, msg_buf+sizeof (sdata), FULL_BANDWIDTH);
	  break;
	case SOUTH:
	  bcopy (brobots_xpm, msg_buf+sizeof (sdata), FULL_BANDWIDTH);
	  break;
	case WEST:
	  bcopy (brobotw_xpm, msg_buf+sizeof (sdata), FULL_BANDWIDTH);
	  break;
	}
      }
      else {  /* Send reduced bandwidth unknown ("partial") image. */
	sdata.position.object_type = OUTSWAP (UNKNOWN);

	sdata.size = OUTSWAP (REDUCED_BANDWIDTH);
	var_data_size = REDUCED_BANDWIDTH;
	bcopy (unknown_xpm, msg_buf+sizeof (sdata), REDUCED_BANDWIDTH);
      }
	
      /* Now copy control info for data message to msg_buf. */
      bcopy (&sdata, msg_buf, sizeof (sdata));

      sendMsg (client[i].conn_id, 
	       msg_buf, 
	       sizeof (sdata) + var_data_size, 
	       "sendEvent");
    }
  }
  
  resync (exchange_list, !resync_flag);
}



void ECendEvent (int (*func)(struct robot_args *), 
		 struct robot_args *arg) {
  
  int i, j;
  int x, y;

  update_time();

  if (func == fire_missile) {
    x = robot_position[id][arg->robot_num].x;
    y = robot_position[id][arg->robot_num].y;
    client_lock[x][y].modified   = FALSE;
    client_lock[x-1][y].modified = FALSE;
    client_lock[x+1][y].modified = FALSE;
    client_lock[x][y+1].modified = FALSE;
    client_lock[x][y-1].modified = FALSE;
    
    sendRelLock (id, &client_lock[x][y]);
    sendRelLock (id, &client_lock[x-1][y]);
    sendRelLock (id, &client_lock[x+1][y]);
    sendRelLock (id, &client_lock[x][y-1]);
    sendRelLock (id, &client_lock[x][y+1]);
  }
  else { 

    /* Event must have been robot walk or rotation. */
    if ((func == rotate_clockwise) ||
	(func == rotate_anticlockwise)) {
      x = robot_position[id][arg->robot_num].x;
      y = robot_position[id][arg->robot_num].y;
      client_lock[x][y].modified = TRUE;
      client_lock[x-1][y].modified = FALSE;
      client_lock[x+1][y].modified = FALSE;
      client_lock[x][y+1].modified = FALSE;
      client_lock[x][y-1].modified = FALSE;
      
      sendRelLock (id, &client_lock[x][y]);
      sendRelLock (id, &client_lock[x-1][y]);
      sendRelLock (id, &client_lock[x+1][y]);
      sendRelLock (id, &client_lock[x][y-1]);
      sendRelLock (id, &client_lock[x][y+1]);
    }
    if (func == walk) {
      /* Send other clients info about the block 
	 from where the robot walked. */
      switch (arg->value) {
      case NORTH:
	x = robot_position[id][arg->robot_num].x;
	y = robot_position[id][arg->robot_num].y+1;
	client_lock[x][y-1].modified = TRUE;
	client_lock[x-1][y].modified = FALSE;
	client_lock[x+1][y].modified = FALSE;
	client_lock[x][y+1].modified = FALSE;
	break;

      case EAST:
	x = robot_position[id][arg->robot_num].x-1;
	y = robot_position[id][arg->robot_num].y;
	client_lock[x+1][y].modified = TRUE;
	client_lock[x-1][y].modified = FALSE;
	client_lock[x][y-1].modified = FALSE;
	client_lock[x][y+1].modified = FALSE;
	break;
	
      case SOUTH:
	x = robot_position[id][arg->robot_num].x;
	y = robot_position[id][arg->robot_num].y-1;
	client_lock[x][y+1].modified = TRUE;
	client_lock[x-1][y].modified = FALSE;
	client_lock[x][y-1].modified = FALSE;
	client_lock[x+1][y].modified = FALSE;
	break;
	
      case WEST:
	x = robot_position[id][arg->robot_num].x+1;
	y = robot_position[id][arg->robot_num].y;
	client_lock[x-1][y].modified = TRUE;
	client_lock[x][y-1].modified = FALSE;
	client_lock[x+1][y].modified = FALSE;
	client_lock[x][y+1].modified = FALSE;
	break;
      }

      client_lock[x][y].modified = TRUE;
      sendRelLock (id, &client_lock[x][y]);
      sendRelLock (id, &client_lock[x-1][y]);
      sendRelLock (id, &client_lock[x+1][y]);
      sendRelLock (id, &client_lock[x][y-1]);
      sendRelLock (id, &client_lock[x][y+1]);

#if 0
      /* This isn't really needed by entry consistency but is used to
	 make the application work "correctly". */

      memset (&sdata, -1, sizeof (sdata));
      sdata.m            = OUTSWAP (data);
      sdata.id           = OUTSWAP (id);
      sdata.ts           = OUTSWAP (current_time);
      sdata.position.orientation = OUTSWAP (0); /* Don't-care value. */
      sdata.position.object_type = OUTSWAP (world[x][y]);
      sdata.position.x = OUTSWAP (x);
      sdata.position.y = OUTSWAP (y);
      sdata.size = 0;		/* No variable data to tag to message. */

      for (i = 0; i < num_clients; i++) {
	if ((client[i].registered == FALSE) || (i == id))
	  continue;
	
	sendMsg (client[i].conn_id, &sdata, 
		 sizeof (sdata), "ECendEvent");
      }
#endif
    }
  }
}


/* Right now the only data we request is ENEMY_ROBOT data 
   in response to seeing an UNKNOWN object type in a specific
   location of world[][]. */
void sendDataReq (datareq_t (*client_list)[MAX_RANGE], 
		  int range) {

  int i, j;

  stext.m    = OUTSWAP (text);
  stext.id   = OUTSWAP (id);
  stext.type = OUTSWAP (DATAREQ);

  for (i = 0; i < num_clients; i++) {
    if ((client[i].registered == FALSE) || (i == id))
      continue;

    for (j = 0; j < range; j++)
      if (client_list[i][j].flag == UNKNOWN) {
	stext.info.datareq.flag = OUTSWAP (ENEMY_ROBOT);
	stext.info.datareq.x    = OUTSWAP (client_list[i][j].x);
	stext.info.datareq.y    = OUTSWAP (client_list[i][j].y);
	
	sendMsg (client[i].conn_id, &stext, 
		 sizeof (stext), "sendDataReq");
	
	/* Wait for requested data to arrive. Check X events 
	   in the meantime. How do we know that socketHandle()
	   has received the requested data and NOT other data
	   from e.g. some client performing a sendEvent()? 
	   What we do is check whether the requesting client's
	   local copy of world[x][y] matches the object type we
	   are requesting. */
	
	while (world[client_list[i][j].x][client_list[i][j].y] !=
	       ENEMY_ROBOT)
	  while (eventCheck (TRUE, &pause_string));
      }
  }
}


void sendToken (int *token, int active_client) {
  int i;
  int next_client = -1;

  printf ("sendToken\n");
  
  /* Pass token to next client in sequence. */
  memset (&stext, -1, sizeof (stext));
  stext.m    = OUTSWAP (text);
  stext.type = OUTSWAP (TOKEN);
  stext.id   = OUTSWAP (id);

  active [id] = active_client;
  bcopy (active, stext.info.token.active, NUM_TEAMS*sizeof(int));
  bcopy (conn_established, stext.info.token.conn_established, 
	 NUM_TEAMS*sizeof(int));

  for (i = 1; i < num_clients; i++)
    if (active[(id+i) % num_clients]) {
      next_client = ((id+i) % num_clients);
      break;
    }

  if (next_client != -1) {

    sendMsg (client[next_client].conn_id, 
	     &stext, sizeof (stext), "sendToken");

    *token = FALSE;
  }

  printf ("next_client = %d, active[%d]=%d,"
	  "conn_established[%d]=%d\n",
	  next_client, id, active[id], id, 
	  conn_established[id]);
}


void applyDataMsg (char *buf, datamsg_t *data) {
  int robot_number;
  int team_number;
  int x, y;
  int orientation;
  int object_type;
  int data_size;
  dlock_t dlock;

  /*printf ("applying\n");*/

  /* Obtain the actual image object and info about its orientation
     and sector. */
  x            = INSWAP (data->position.x);
  y            = INSWAP (data->position.y);
  orientation  = INSWAP (data->position.orientation);
  object_type  = INSWAP (data->position.object_type);
  
  world [x][y] = object_type;

  if ((INSWAP (data->m)) == setup) {
    bcopy (buf, &dlock, sizeof(dlock_t));
    free (buf);
    buf = NULL;

    client_lock[x][y].mgr       = INSWAP (dlock.mgr);
    client_lock[x][y].lock_id   = INSWAP (dlock.lock_id);
    mgr_lock[x][y].mgr       = INSWAP (dlock.mgr);
    mgr_lock[x][y].lock_id   = INSWAP (dlock.lock_id);
    lock_tt[INSWAP (dlock.lock_id)].x = x;
    lock_tt[INSWAP (dlock.lock_id)].y = y;
  }

  /* Even though we have an object type of UNKNOWN, it must actually
     be a robot of some sort. Right now, it is merely an ENEMY_ROBOT,
     but for our intents and purposes we assume no such knowledge. */
  if (object_type == FRIENDLY_ROBOT || 
      object_type == ENEMY_ROBOT ||
      object_type == UNKNOWN) { 
    /* Setup local data structures. */
    team_number = INSWAP (data->id);
    client[team_number].team_size = INSWAP (data->rinfo.team_size);
    robot_number = INSWAP (data->rinfo.robot_number);
    robot_position [team_number][robot_number].x = x;
    robot_position [team_number][robot_number].y = y;
    robot_front [team_number][robot_number] = orientation;
    robot_life [team_number][robot_number] = 
      INSWAP (data->rinfo.robot_life);
    killed_neg [team_number][robot_number] = 
      INSWAP (data->rinfo.killed_neg);
    
    if (robot_life [team_number][robot_number] > 0) {
	drawObject (sector, x, y, orientation, object_type, buf);
    }
    else {
      wipeObject (sector, x, y);
      if (killed_neg[team_number][robot_number]) {
	killed_neg[team_number][robot_number] = FALSE;
	/*	drawNegObs (sector, x, y);
	world[x][y] = NEG_OBS;*/
	world[x][y] = UNCHARTED; /* ?? Test */
      }
      else
	world[x][y] = UNCHARTED;

      /* Invalidate robot_position. */
      robot_position[team_number][robot_number].x = -1;
      robot_position[team_number][robot_number].y = -1; 
    }
    
    if (equal (robot_position [team_number][robot_number], goal)) {
      glorySequence (sector, goal.x, goal.y, ENEMY_ROBOT);
      redrawGoal (sector, goal);
      world [goal.x][goal.y] = UNCHARTED;

      /* These next two lines are needed so that getRobotInfo() does't
	 incorrectly get the wrong robot in the future. */
      robot_position[team_number][robot_number].x = -1;
      robot_position[team_number][robot_number].y = -1; 

#ifdef SOUND
#ifdef SOUNDCARD
      if (sound)
	playSoundFile ("goal");
#endif
#endif
    }
  }
  else {
    if (object_type == GOAL) {
      goal.x = x;
      goal.y = y;
    }
    wipeObject (sector, x, y);
    drawObject (sector, x, y, orientation, object_type, buf);
  }
  
  if (buf) {
    free (buf);
    buf = NULL;
  }
}


void bufferDataMsg (char *msg_buf, int slot_id) {

  int var_data_size;
  databuf_t *bucket;
  databuf_t *buf_ptr;

  /*printf ("buffering\n");*/

  if (databuf[slot_id].status == occupied) {
    bucket = (databuf_t *) malloc (sizeof(databuf_t));
    bucket->status = occupied;
    bcopy (msg_buf, 
	   &bucket->m, 
	   sizeof(datamsg_t));

    if ((var_data_size = INSWAP (bucket->m.size)) > 0) {
      bucket->buf = (char *) malloc (var_data_size);
      bcopy (msg_buf + sizeof(datamsg_t), 
	     bucket->buf,
	     var_data_size);
    }
    buf_ptr = &databuf[slot_id];
    while (buf_ptr->next)
      buf_ptr = buf_ptr->next;
    buf_ptr->next = bucket;
    bucket->next = NULL;
  }
  else {
    databuf[slot_id].status = occupied;
    bcopy (msg_buf, 
	   &databuf[slot_id].m, 
	   sizeof(datamsg_t));

    if ((var_data_size = INSWAP (databuf[slot_id].m.size)) > 0) {
      databuf[slot_id].buf = (char *) malloc (var_data_size);
      bcopy (msg_buf + sizeof(datamsg_t), 
	     databuf[slot_id].buf,
	     var_data_size);
    }
    databuf[slot_id].next = NULL; /* Currently no chained buckets. */
  }
}


void retrieveDataMsg (int slot_id) {

  int var_data_size;
  databuf_t *buf_ptr;
 
  /*printf ("retrieving\n");*/

  buf_ptr = &databuf[slot_id];
  
  while (buf_ptr && 
	 (buf_ptr->status == occupied) &&
	 ((INSWAP (buf_ptr->m.ts)) <= current_time)) { 
    /* Retrieve and apply to mem. */
    if ((var_data_size = INSWAP (buf_ptr->m.size)) > 0)
      applyDataMsg (buf_ptr->buf, &buf_ptr->m);
    else
      applyDataMsg (NULL, &buf_ptr->m);
  
    if (buf_ptr != (&databuf[slot_id])) {
      databuf[slot_id].next = buf_ptr->next;
      free (buf_ptr);
    }
    buf_ptr = databuf[slot_id].next;

    if (databuf[slot_id].next == NULL)
      databuf[slot_id].status = empty;  
  }
}


/* Handle data from network arriving via a socket. */
int socketHandle (int timeout) {
  int robot_number;
  int team_number;
  int x, y;
  struct robot_args args;
  msg_t msg_type;
  int i;
  char *buf=NULL;
  int var_data_size;
  dlock_t dlock;

  if (Net_Poll (timeout, msg_buf, &msg_len, &rconn_id) != NET_SUCCESS)
    return (FALSE);

  bcopy (msg_buf, (char *) &msg_type, sizeof(msg_t));
  msg_type = INSWAP (msg_type);

  if ((msg_type == data) || (msg_type == setup)) {
    /* Copy data message to rdata structure. */
    bcopy (msg_buf, (char *) &rdata, sizeof(datamsg_t));
    if ((INSWAP (rdata.id)) == -1)
      return (SETUP_COMPLETE);

    if ((INSWAP (rdata.ts)) > current_time)
      bufferDataMsg (msg_buf, INSWAP (rdata.id));
    else {
      var_data_size = INSWAP (rdata.size);
      if (var_data_size > 0) {
	buf = (char *) malloc (var_data_size);
	bcopy (msg_buf + sizeof(datamsg_t), buf, var_data_size);
      }
      applyDataMsg (buf, &rdata);
    }
    return (DATAMSG);
  }
  
  else if (msg_type == text) {
    /* Copy text message to rtext structure. */
    bcopy (msg_buf, (char *) &rtext, msg_len);

    switch (INSWAP (rtext.type)) {

    case SIGNOFF:
      printf ("%s: Signoff from client %d.\n", progname, id);
      printf ("Reason: %s\n", rtext.info.signoff.reason);
      stop = TRUE;
      return (SIGNOFF);
      
    case ACCEPT:
      id          = INSWAP (rtext.info.accept.id);
      goal_score  = INSWAP (rtext.info.accept.goal_score);
      num_clients = INSWAP (rtext.info.accept.num_clients);
      for (i = 0; i < num_clients; i++) {
	client[i].port       = INSWAP (rtext.info.accept.client[i].port);
	client[i].registered = INSWAP (rtext.info.accept.client[i].registered);
	strcpy (client[i].username, rtext.info.accept.client[i].username);
	strcpy (client[i].hostname, rtext.info.accept.client[i].hostname);
      }
      con_protocol = INSWAP (rtext.info.accept.cp);
      
      /* Set list of active clients to TRUE at start of game. */
      for (i = 0; i < num_clients; i++)
	active[i] = TRUE;
      
      return (ACCEPT);

    case NEGOTIATE:
      return (NEGOTIATE);

    case REJECT:
      fprintf (stderr, "Signon rejected by server.\n");
      fprintf (stderr, "Reason: %s\n", rtext.info.reject.text.text_info);
      shutdown (1);
      return (REJECT);

    case DATAREQ:
      x = INSWAP (rtext.info.datareq.x);
      y = INSWAP (rtext.info.datareq.y);

      memset (&sdata, -1, sizeof (sdata));
      sdata.m            = OUTSWAP (data);
      sdata.id           = OUTSWAP (id);
      sdata.ts           = OUTSWAP (0);

      if ((world[x][y] == ENEMY_ROBOT) ||
	  (world[x][y] == FRIENDLY_ROBOT) ||
	  (world[x][y] == UNKNOWN)) {
	/* For a valid data request, the following team_number 
	   should match this client's id.*/
	getRobotInfo (x, y, &team_number, &robot_number);
	
	sdata.rinfo.team_size      = OUTSWAP (client[team_number].team_size);
	sdata.rinfo.robot_number   = OUTSWAP (robot_number);
	sdata.rinfo.robot_life     = 
	  OUTSWAP (robot_life[team_number][robot_number]);
	sdata.rinfo.killed_neg     = 
	  OUTSWAP (killed_neg[team_number][robot_number]);
	sdata.position.x   = 
	  OUTSWAP (robot_position[team_number][robot_number].x);
	sdata.position.y   = 
	  OUTSWAP (robot_position[team_number][robot_number].y);
	sdata.position.orientation = 
	  OUTSWAP (robot_front[team_number][robot_number]);
	
	if (team_number == id)
	  sdata.position.object_type = OUTSWAP (ENEMY_ROBOT);
	else
	  sdata.position.object_type = OUTSWAP (FRIENDLY_ROBOT);

	sdata.size = OUTSWAP (FULL_BANDWIDTH);

	switch (robot_front[team_number][robot_number]) {
	case NORTH:
	  bcopy (brobotn_xpm, msg_buf+sizeof (sdata), FULL_BANDWIDTH);
	  break;
	case EAST:
	  bcopy (brobote_xpm, msg_buf+sizeof (sdata), FULL_BANDWIDTH);
	  break;
	case SOUTH:
	  bcopy (brobots_xpm, msg_buf+sizeof (sdata), FULL_BANDWIDTH);
	  break;
	case WEST:
	  bcopy (brobotw_xpm, msg_buf+sizeof (sdata), FULL_BANDWIDTH);
	  break;
	}
	  
	/* Now copy control info for data message to msg_buf. */
	bcopy (&sdata, msg_buf, sizeof (sdata));
	  
	sendMsg (client[INSWAP (rtext.id)].conn_id, 
		 msg_buf, 
		 sizeof (sdata) + FULL_BANDWIDTH, 
		 "socketHandle");
      }
      else {
	sdata.size                 = OUTSWAP (0);/* No variable-length data.*/
	sdata.position.x           = OUTSWAP (x);
	sdata.position.y           = OUTSWAP (y);
	sdata.position.object_type = OUTSWAP (world[x][y]);
	sdata.position.orientation = OUTSWAP (0); /* Don't-care value. */
	sendMsg (client[INSWAP (rtext.id)].conn_id, 
		 &sdata, 
		 sizeof (sdata),
		 "socketHandle");
      }
      return (DATAREQ);

    case SCORE:
      team_number = INSWAP (rtext.id);
      goal_score  = INSWAP (rtext.info.score.goal_score);
      cumulative_tot_score [team_number] = 
	INSWAP (rtext.info.score.cumulative_tot);
      session_tot_score [team_number] = 
	INSWAP (rtext.info.score.session_tot);
      
      refreshSWindow (cumulative_tot_score, session_tot_score);
      return (SCORE);

    case A_MISSILE:
      if (INSWAP (rtext.info.missile.active)) {
	/* We have a missile to process. */
	robot_number   = INSWAP (rtext.info.missile.robot_number);
	team_number    = INSWAP (rtext.id);
	args.robot_num = robot_number;
	args.value     = team_number;
	args.x         = INSWAP (rtext.info.missile.x);
	args.y         = INSWAP (rtext.info.missile.y);
	args.orientation = INSWAP (rtext.info.missile.orientation);
	fire_missile (&args);
      }
      return (A_MISSILE);

      /* This client should only receive a token if it is still
         playing in an active game. */
    case TOKEN:
      free_token = TRUE;
      
      bcopy (rtext.info.token.conn_established, conn_established, 
	     NUM_TEAMS*sizeof(int));
      /* Update list of active clients still playing current game. */
      bcopy (rtext.info.token.active, active, NUM_TEAMS*sizeof(int));
      return (TOKEN);
      
    case SYNC:
      sync_list [INSWAP (rtext.id)].sync_flag = TRUE;
      sync_list [INSWAP (rtext.id)].pending++;
      sync_list [INSWAP (rtext.id)].resync_flag = 
	INSWAP (rtext.info.sync.resync_flag);
      sync_list [INSWAP (rtext.id)].ts = 
	INSWAP (rtext.info.sync.ts);
      free_data_token = TRUE;
      return (SYNC);

    case CLIENT_LK_ACQ:
      dlock.lock_id  = INSWAP (rtext.info.dlock.lock_id);
      dlock.modified = INSWAP (rtext.info.dlock.modified);
      dlock.client_id= INSWAP (rtext.info.dlock.client_id);

      x = lock_tt[dlock.lock_id].x;
      y = lock_tt[dlock.lock_id].y;
      client_lock[x][y].modified  = dlock.modified;
      client_lock[x][y].client_id = dlock.client_id;
      if (client_lock[x][y].type == LK_WRITE)
	client_lock[x][y].status = locked;
      /* ?? lock_granted = TRUE; */

#ifdef DEBUG
      printf ("(socklink) Rx'ed CLIENT_LK_ACQ\n");
#endif
      return (CLIENT_LK_ACQ);

    case MGR_LK_ACQ:
      dlock.lock_id   = INSWAP (rtext.info.dlock.lock_id);
      dlock.type      = INSWAP (rtext.info.dlock.type);
      dlock.ts        = INSWAP (rtext.info.dlock.ts);
      dlock.client_id = INSWAP (rtext.info.dlock.client_id);

#ifdef DEBUG
      printf ("(socklink) Rx'ed MGR_LK_ACQ:"
	      "Lock id %d, type %d, ts %d, client_id %d\n",
	      dlock.lock_id, dlock.type, dlock.ts, dlock.client_id);      
#endif

      acquireLockHandler (INSWAP (rtext.id), TRUE, &dlock);
      return (MGR_LK_ACQ);

    case MGR_LK_REL:
      dlock.lock_id  = INSWAP (rtext.info.dlock.lock_id);
      dlock.ts       = INSWAP (rtext.info.dlock.ts);
      dlock.modified = INSWAP (rtext.info.dlock.modified);
      dlock.client_id= INSWAP (rtext.info.dlock.client_id);
      dlock.type     = INSWAP (rtext.info.dlock.type);

#ifdef DEBUG
      printf ("(socklink) Rx'ed MGR_LK_REL:"
	      "Lock id %d, type %d, ts %d, client_id %d, modified %d\n",
	      dlock.lock_id, dlock.type, dlock.ts, 
	      dlock.client_id, dlock.modified);      
#endif

      releaseLockHandler (INSWAP (rtext.id), &dlock);
      return (MGR_LK_REL);

    default:
      fprintf (stderr, "%s: Unexpected message type %d.\n", 
	       progname, INSWAP (rtext.type));
      return (FALSE);
    }
  }

  else {
    fprintf (stderr, "%s: Incorrect message format\n", progname);
    shutdown (1);
  }
}


int ECgetData (int client_id, int x, int y) {
  
  static int first_try = TRUE;

  if (first_try) {
    first_try = FALSE;
    
    stext.m                 = OUTSWAP (text);
    stext.id                = OUTSWAP (id);
    stext.type              = OUTSWAP (DATAREQ);
    stext.info.datareq.flag = OUTSWAP (0);
    stext.info.datareq.x    = OUTSWAP (x);
    stext.info.datareq.y    = OUTSWAP (y);
    
    sendMsg (client[client_id].conn_id, 
	     &stext, 
	     sizeof (stext), 
	     "ECgetData");
  }

  if (socketHandle (0) == DATAMSG) {
    first_try = TRUE;
    return (DATAMSG);
  }
  else
    return (FALSE);
}


/* ?? Beware that network-wide data races can occur with the
 scoring as implemented below. The 'goal_score' should be a lockable
 shared resource, since the network latency can delay the newest
 value of 'goal_score' reaching a client. During the wire-time,
 the local client could reach the goal and appear to have a better
 score than allowed by using the old 'goal_score' value. */
void update_scores (int team_number, int robot_number) {
  
  int i;

  /* Adjust scores when a team's robot reaches the goal. */
  cumulative_tot_score [team_number] += 
    (10 * ((float)(goal_score * robot_life[team_number][robot_number]) /
     (team_size*2)));
  session_tot_score [team_number]    += 
    (10 * ((float)(goal_score-- * robot_life[team_number][robot_number]) / 
     (team_size*2)));

  /* Broadcast new scores to all active clients.
     Only need to broadcast team scores that have changed. */
  memset (&stext, -1, sizeof (stext));
  stext.m    = OUTSWAP (text);
  stext.type = OUTSWAP (SCORE);
  /* Cunningly use the id member to specify the team to which
     the score corresponds. */
  stext.id   = OUTSWAP (team_number); 
  stext.info.score.goal_score = 
    OUTSWAP (goal_score);
  stext.info.score.session_tot = 
    OUTSWAP (session_tot_score [team_number]);
  stext.info.score.cumulative_tot = 
    OUTSWAP (cumulative_tot_score [team_number]);

  for (i = 0; i < num_clients; i++) {
    if ((client[i].registered == FALSE) || (i == id))
      continue;
    
    sendMsg (client[i].conn_id, &stext, sizeof (stext), "update_scores");
  }
  refreshSWindow (cumulative_tot_score, session_tot_score);
}












