#include "Xdata.h"
#include "consistency.h"
#include "tk.h"

char *pause_string = "PAUSE";
int cumulative_tot_score [NUM_TEAMS];
int session_tot_score [NUM_TEAMS];

extern databuf_t databuf[NUM_TEAMS];
extern int paused;
extern int stop;
extern int end_prog;
extern int sched_num;
extern int team_size;
extern char *sector; 
extern char *current_sector; 
extern char *opmode; 
extern client_t client [NUM_TEAMS];
extern unsigned long current_time;
extern int conn_established[NUM_TEAMS];

extern void refreshSWindow (int *cumulative_tot_score, 
			    int *session_tot_score);
extern void getRobotInfo (int x, int y, 
			  int *team_number, 
			  int *robot_number);
extern char *getActiveSector (int x, int y);

int eventCheck (int flag, char **string) {
  int width, height, code;
  char geometry [10];
  int x, y, offset_x, offset_y;
  static int robot_id=-1; /* User mode only. */
  int team_id;
  coordinate coord;
  char message [MSG_LEN];
  XEvent key_event;
  KeySym keysym;
  static int rotation = CLOCKWISE;

  if (end_prog) {
    XCloseDisplay (Disp);
    sprintf (message, "User %d has quit game.", id);
    signoff (S_QUIT, message);
  }

  if (flag)
    socketHandle (0);

  while (XCheckTypedEvent (Disp, Expose, &report)) {
    if (report.xexpose.count == 0) {

      /* Keep Tcl/Tk vars consistent, 
	 e.g. Tcl "sector" variable consistent with C "sector" variable. */
      sector = Tcl_GetVar (Et_Interp, "sector", TCL_GLOBAL_ONLY);
      getSectorOffsets (sector, &offset_x, &offset_y);
      refreshWorld (current_sector, sector, offset_x, offset_y, MAXX, MAXY);
      refreshCWindows (*string);
    }
  }

  while (XCheckTypedEvent (Disp, ButtonPress, &report)) {

    if ((report.xbutton.window == FWin) &&
	(strcmp (opmode, "User") == 0)) {
      /* Get battlefield block where mouse pointer is located. */
      x = (int)((double)report.xbutton.x / IMAGEX);
      y = (int)((double)report.xbutton.y / IMAGEY);
      getSectorOffsets (sector, &offset_x, &offset_y);
      if (world [x+offset_x][y+offset_y] == FRIENDLY_ROBOT) {
	coord.x = x+offset_x;
	coord.y = y+offset_y;
	getRobotInfo (coord.x, coord.y, &team_id, &robot_id);
      }
      if (robot_id >= 0)
	createTaskFromInput (sched_num, 
			     robot_id, 
			     report.xbutton.button,
			     rotation);
    }

    /* Disable button presses on START/STOP/PAUSE buttons if flag is FALSE. */
    else if ((report.xbutton.window == CWin2)  && flag) { /* Stop. */
      XDrawRectangle (Disp, CWin2, GcDF, 0, 0, 
		      CONTROL_WIDTH - 1, CONTROL_HEIGHT - 1); 
      sprintf (message, "User %d has stopped game.", id);
      signoff (S_STOP, message);
    }

    else if ((report.xbutton.window == CWin3) && 
	     (report.xbutton.button == Button1) && flag) {
      XClearWindow (Disp,CWin3);
      XDrawRectangle (Disp, CWin3, GcDF, 0, 0, 
		      2*CONTROL_WIDTH + 2*BORDER - 1, 
		      CONTROL_HEIGHT - 1); 
      *string = "PAUSED";
      XDrawString (Disp,CWin3,GcDF,60,25,
		   *string,strlen(*string));
      paused = TRUE;
    }

    else if ((report.xbutton.window == CWin3) && 
	     (report.xbutton.button == Button3) && flag) {
      XClearWindow (Disp,CWin3);
      *string = "PAUSE";
      XDrawString (Disp,CWin3,GcDF,60,25,
		   *string,strlen(*string));
      paused = FALSE;
    }

    else if (report.xbutton.window == RWin) {

      sector = getActiveSector (report.xbutton.x, report.xbutton.y); 
      
      if (strcmp (current_sector, sector) != 0) {
	getSectorOffsets (sector, &offset_x, &offset_y);
	refreshWorld (current_sector, 
		      sector, 
		      offset_x, 
		      offset_y, 
		      MAXX, 
		      MAXY);
	
	/* Keep Tcl/Tk "sector" variable consistent 
	   with C "sector" variable. */
	Tcl_SetVar (Et_Interp, "sector", sector, TCL_GLOBAL_ONLY);
      }
    }
  }
      
  /* Check for spacebar keypress to alter direction of robot rotation. */
  if (XCheckTypedEvent (Disp, KeyPress, &report)) {
    keysym = XLookupKeysym (&report.xkey, 0);

    if (keysym == XK_space)
      rotation = (rotation == CLOCKWISE) ? 
	ANTICLOCKWISE : CLOCKWISE;

    else			/* Event might be for Tk widgets. */
      XPutBackEvent (Disp, &report);
  }

  while (XCheckTypedEvent (Disp, ConfigureNotify, &report)) {
    width = MAXX*IMAGEX+STATUS_WIDTH+6*BORDER;
    height = MAXY*IMAGEY+4*BORDER+MENU_HEIGHT;

    sprintf (geometry, "%d%c%d", width, 'x', height);
    
    Tcl_SetVar (Et_Interp, "geometry", 
		geometry, TCL_GLOBAL_ONLY);

    code = Tcl_VarEval (Et_Interp, "wm geometry . ", 
			geometry, (char *) NULL);
    
    if (code != TCL_OK)
      fprintf(stderr, "%s\n", Et_Interp->result);
  }
  
  /* Handle Tk events. */
  Tcl_Eval (Et_Interp, "update");

  /* Discard any other events in queue. Be careful not to discard potentially
     useful events in the client's request buffer so use XQLength() instead
     of XPending(). */
  while (XQLength (Disp))
    XNextEvent (Disp, &report);

  XFlush (Disp);
  return (paused);
}


void eventHandle (int port) {
  int i, j, width, height, code, popped = FALSE;
  char geometry [10];
  int start_game = TRUE;
  int offset_x, offset_y;

  while (TRUE)  { 
    if (end_prog)
      signoff (S_QUIT, "User has quit game.");
    
    socketHandle (0);

    if (popped && (team_size == -1)) 
      /* User has yet to input number of robots per team into 
	 displayed dialog box. */
      ET( DialogBox );
    
    /* User must have input via the popup dialog box. */
    if (team_size > 0) { 


      /* Ensure Tk widget-events are serviced. */
      Tcl_Eval (Et_Interp, "update"); 
      drawBackdrop (current_sector, 0, 0, MAXX*IMAGEX, MAXY*IMAGEY);

      popped = FALSE;
      current_time = 0;
      zap_world ();
      for (i = 0; i < NUM_TEAMS; i++) {
	databuf[i].status = empty;
	for (j = 0; j < MAX_ROBOTS; j++) {
	  robot_position[i][j].x = -1;
	  robot_position[i][j].y = -1; 
	  robot_front[i][j]  = 0; 
	  robot_life[i][j]   = 0; 
	}
      }
      memset (session_tot_score, 0, sizeof (int)*NUM_TEAMS);
      memset (cumulative_tot_score, 0, sizeof (int)*NUM_TEAMS);
      paused = FALSE;
      bzero (conn_established, NUM_TEAMS*sizeof(int));
      initLocks ();
      /* Flush connections. */
      while (socketHandle (0))
	while (eventCheck (FALSE, &pause_string));

      /* Sign on with server before proceeding. */
      signon (team_size, port);

      /* Setup Consistency Protocol. */
      setupConProto (con_protocol);

#if 0
      {
        int i, j;
        extern dlock_t client_lock[FIELDX][FIELDY];
        extern coordinate lock_tt[FIELDX*FIELDY];

	for (j = 0; j < FIELDY; j++)	
	  for (i = 0; i < FIELDX; i++)
            printf ("client_lock[%d][%d].mgr = %d\t"
                    "client_lock[%d][%d].lock_id = %d\t"
                    "lock_tt[%d] = (%d,%d)\n", 
                    i, j, client_lock[i][j].mgr, 
                    i, j, client_lock[i][j].lock_id,
                    j*FIELDX+i, 
                    lock_tt[j*FIELDX+i].x,
                    lock_tt[j*FIELDX+i].y);
      }
#endif

      if (start_game)
	while (negotiate (id) == FALSE)
	  while (eventCheck (FALSE, &pause_string));

      start_game = FALSE;

      drawBackdrop (current_sector, 
		    0, 0, MAXX*IMAGEX, MAXY*IMAGEY);

      getSectorOffsets (sector, 
			&offset_x, 
			&offset_y);

      refreshWorld (current_sector, 
		    sector, 
		    offset_x, 
		    offset_y, 
		    MAXX, 
		    MAXY);

      init_buffers_and_queues ();
      memset (session_tot_score, 0.0, sizeof (float)*NUM_TEAMS);
      refreshSWindow (cumulative_tot_score, session_tot_score);
      stop = FALSE;

      opmode = Tcl_GetVar (Et_Interp, "opmode", TCL_GLOBAL_ONLY);

      /* Long process, so call "eventCheck" every so often. */
      robotControl (opmode);

      /* Reset team_size to invalid state. */
      team_size = -1;  
      current_time = 0;
    }
      
    while (XCheckTypedEvent (Disp, Expose, &report)) {
      if (report.xexpose.count == 0) {

	if (report.xexpose.window == FWin) {

	  /* Keep Tcl/Tk "sector" variable consistent 
	     with C "sector" variable. */
	  sector = Tcl_GetVar (Et_Interp, "sector", TCL_GLOBAL_ONLY);
	  getSectorOffsets (sector, &offset_x, &offset_y);
	  refreshWorld (current_sector, 
			sector, 
			offset_x, 
			offset_y, 
			MAXX, 
			MAXY);
	}
	
	refreshCWindows (pause_string);
      }
    }
    
    while (XCheckTypedEvent (Disp, ButtonPress, &report)) {

      /* Start window button. */
      if ((report.xbutton.window == CWin1) 
	  && (report.xbutton.button == Button1)) {
	XDrawRectangle (Disp, CWin1, GcDF, 0, 0, 
			CONTROL_WIDTH - 1, CONTROL_HEIGHT - 1); 
	if (team_size <= 0) {
	  /* No robots yet defined, so raise a popup dialog box. */
	  popped = TRUE;
	  team_size = -1;
	}
      }
      
      else if (report.xbutton.window == CWin2) {
	XDrawRectangle (Disp, CWin2, GcDF, 0, 0, 
			CONTROL_WIDTH - 1, CONTROL_HEIGHT - 1); 
	stop = TRUE;
      }
      
      else if (report.xbutton.window == RWin) {
	sector = getActiveSector (report.xbutton.x, report.xbutton.y); 
	if (strcmp (current_sector, sector) != 0) {
	  getSectorOffsets (sector, &offset_x, &offset_y);
	  refreshWorld (current_sector, 
			sector, 
			offset_x, 
			offset_y, 
			MAXX, 
			MAXY);
	  
	  /* Keep Tcl/Tk "sector" variable consistent 
	     with C "sector" variable. */
	  Tcl_SetVar (Et_Interp, "sector", sector, TCL_GLOBAL_ONLY);
	}
      }
    }
    
    while (XCheckTypedEvent (Disp, ConfigureNotify, &report)) {
      width = MAXX*IMAGEX+STATUS_WIDTH+6*BORDER;
      height = MAXY*IMAGEY+4*BORDER+MENU_HEIGHT;
      
      sprintf (geometry, "%d%c%d", width, 'x', height);
      
      Tcl_SetVar (Et_Interp, "geometry", 
		  geometry, TCL_GLOBAL_ONLY);
      code = Tcl_VarEval (Et_Interp, "wm geometry . ", 
			  geometry, (char *) NULL);
      
      if (code != TCL_OK)
	fprintf(stderr, "%s\n", Et_Interp->result);
    }
    
    Tcl_Eval (Et_Interp, "update"); 
    /* Ensure Tk widget-events are serviced. */
  }
}








