#include "Xdata.h"
#include "comms.h"
#include "tk.h"
#include <signal.h>

extern char *progname;
extern int robot_delay;
extern int sound;
extern int cumulative_tot_score [NUM_TEAMS];
extern int session_tot_score [NUM_TEAMS];
extern char *sector; 
extern char *current_sector; 
extern unsigned long current_time;
extern int child_pid, parent_pid;
extern client_t client [NUM_TEAMS];
extern int num_clients;

extern void redrawGoal (char *sector, coordinate goal);
extern void getRobotInfo (int x, int y, 
			  int *team_number, 
			  int *robot_number);

static unsigned long last_time;
int activeSector (char *sector, int x, int y);

void shutdown (int exit_num) {
  int i;

  freeAudioSystem ();

  if (getpid() == child_pid)
    kill (parent_pid, SIGKILL);
  else
    kill (child_pid, SIGKILL);

  exit (exit_num);
}


unsigned long getColor (char name[]) {
  XColor tmp;

  if (XParseColor(Disp, colormap, name, &tmp) == 0) {
    (void) fprintf(stderr, "Battle Robots: Invalid color '%s'.\n", name);
    return BlackPixel(Disp, Scr);
  }
  if (XAllocColor(Disp, colormap, &tmp) == 0) {
    (void) fprintf(stderr, 
		   "Battle Robots: Can't allocate color '%s'.\n", name);
    return BlackPixel(Disp, Scr);
  }
  return tmp.pixel;
}


void initColors () {
  int i;
  
  for (i = 0; i < NUM_OF_COLORS; i++) {
    colorPixelValues[i] = getColor(colorNames[i]);
  }
}


/* Refresh Control Windows (Buttons). */
void refreshCWindows (char *string) {
  char *str1="START";
  char *str2="STOP";
  XClearWindow (Disp,CWin1);
  XClearWindow (Disp,CWin2);
  XClearWindow (Disp,CWin3);
  XDrawString (Disp,CWin1,GcDF,23,25,str1,strlen(str1));
  XDrawString (Disp,CWin2,GcDF,26,25,str2,strlen(str2));
  if (strcmp (string, "PAUSED") == 0)
    XDrawRectangle (Disp, CWin3, GcDF, 0, 0, 
		    2*CONTROL_WIDTH + 2*BORDER - 1, CONTROL_HEIGHT - 1); 
  XDrawString (Disp,CWin3,GcDF,60,25,string,strlen(string));
  XFlush (Disp);
}


void refreshSWindow (int *cumulative_tot_score, 
		     int *session_tot_score) {

  char str[80];
  char substring[80];
  int i;

  for (i = 0; i < num_clients; i++) {

    bzero (str, 80);
    bzero (substring, 80);
  
    sprintf (str, "%s%c%s\t%d\t%d", 
	     client[i].username,
	     '@',
	     client[i].hostname,
	     session_tot_score[i],
	     cumulative_tot_score[i]);
    
    sprintf (substring, "%s%c%s", 
	     client[i].username,
	     '@',
	     client[i].hostname);
    
    ET( RefreshStatusWindow .l "%q(str)" "%q(substring)" );
  }
}


void getSectorOffsets (char *sector, int *init_x, int *init_y) {

  if (strcmp (sector, "Top Left") == 0) {
    *init_x = 0;
    *init_y = 0;
  }
  
  else
    if (strcmp (sector, "Top Right") == 0) {
      *init_x = MAXX;
      *init_y = 0;
    }
  
    else
      if (strcmp (sector, "Bottom Left") == 0) {
	*init_x = 0;
	*init_y = MAXY;
      }
  
      else
	if (strcmp (sector, "Bottom Right") == 0) {
	  *init_x = MAXX;
	  *init_y = MAXY;
	}
}


void refreshRadar (char *sector) {
  int tmpx, tmpy;
  coordinate coord;
  int robot_id, missile_id;
  int init_x, init_y;

  XClearWindow (Disp, RWin);

  getSectorOffsets (sector, &init_x, &init_y);
  XFillRectangle (Disp, RWin, GcSector, init_x*RADAR_X,
		  init_y*RADAR_Y, RADAR_WIDTH*.5, RADAR_HEIGHT*.5);

  for (tmpx = 0; tmpx < FIELDX; tmpx++)
    for (tmpy = 0; tmpy < FIELDY; tmpy++)
      switch (world [tmpx][tmpy]) {
      case FRIENDLY_ROBOT:
	XFillRectangle (Disp, RWin, GcGRF, tmpx*RADAR_X, 
			tmpy*RADAR_Y, RADAR_X, RADAR_Y);
	break;

      case ENEMY_ROBOT:
      case UNKNOWN:
	XFillRectangle (Disp, RWin, GcBRF, tmpx*RADAR_X, 
			tmpy*RADAR_Y, RADAR_X, RADAR_Y);
	break;

      case POS_OBS:
	XDrawRectangle (Disp, RWin, GcPF, tmpx*RADAR_X,
			tmpy*RADAR_Y, RADAR_X-1, RADAR_Y-1);
	break;
	
      case NEG_OBS:
	XDrawRectangle (Disp, RWin, GcNF, tmpx*RADAR_X, 
			tmpy*RADAR_Y, RADAR_X-1, RADAR_Y-1);
	break;

      case STATIC_OBS:
	XDrawRectangle (Disp, RWin, GcDF, tmpx*RADAR_X, 
			tmpy*RADAR_Y, RADAR_X-1, RADAR_Y-1);
	break;

      case N_MISSILE:
      case E_MISSILE:
      case S_MISSILE:
      case W_MISSILE:
	break;

      default:
	coord.x = tmpx;
	coord.y = tmpy;
	if (equal (coord, goal))
	  XFillRectangle (Disp, RWin, GcGF, tmpx*RADAR_X, 
			  tmpy*RADAR_Y, RADAR_X, RADAR_Y);
	break;
      }
}


void drawLifePoints (char *sector,
		     int x,
		     int y,
		     int team_number,
		     int robot_number,
		     int orientation) {

  int i;
  int clipx_offset, clipy_offset;
  int init_x, init_y;

  if (!activeSector (sector, x, y))
    return;

  getSectorOffsets (sector, &init_x, &init_y);
  i = robot_life [team_number][robot_number];
  
  switch (orientation) {
  case NORTH:
    clipx_offset = 12;
    clipy_offset = 15;
    break;
  case EAST:
    clipx_offset = 9;
    clipy_offset = 12;
    break;
  case SOUTH:
    clipx_offset = 12;
    clipy_offset = 9;
    break;
  case WEST:
    clipx_offset = 15;
    clipy_offset = 12;
    break;
  }
	  
  XSetClipOrigin (Disp, 
		  GcNum, 
		  clipx_offset+(x-init_x)*IMAGEX,
		  clipy_offset+(y-init_y)*IMAGEY);

  XCopyArea (Disp, 
	     num[i-1], 
	     FWin, 
	     GcNum, 
	     0, 
	     0, 
	     16, 
	     16, 
	     clipx_offset+(x-init_x)*IMAGEX,
	     clipy_offset+(y-init_y)*IMAGEY);
  
  XFlush (Disp);
}


void drawBackdrop (char *current_sector, 
		   int x, 
		   int y, 
		   int width, 
		   int height) {

  if (strcmp (current_sector, "Top Left") == 0)
    XFillRectangle (Disp, FWin, GcBattle[0], x, y, width, height);
  else
    if (strcmp (current_sector, "Top Right") == 0)
      XFillRectangle (Disp, FWin, GcBattle[1], x, y, width, height);
    else
      if (strcmp (current_sector, "Bottom Right") == 0)
	XFillRectangle (Disp, FWin, GcBattle[2], x, y, width, height);
      else
	if (strcmp (current_sector, "Bottom Left") == 0)
	  XFillRectangle (Disp, FWin, GcBattle[3], x, y, width, height);

  XFlush (Disp);
}


void refreshWorld (char *current_sector, 
		   char *new_sector,
		   int x,
		   int y,
		   int width,
		   int height) {

  int tmpx, tmpy, init_x, init_y;
  coordinate coord;
  int team_id, robot_id, missile_id;

  getSectorOffsets (new_sector, &init_x, &init_y);

  if (strcmp (new_sector, current_sector) != 0) {
    strcpy (current_sector, new_sector);
    ET( UpdateSectorInfo "%s(current_sector)" );

#ifdef SOUND
#ifdef SOUNDCARD
    if (sound)
      playSoundFile ("sector");
#endif
#endif
  }

  refreshRadar (new_sector);
  drawBackdrop (current_sector, 
		(x-init_x)*IMAGEX, 
		(y-init_y)*IMAGEY, 
		width*IMAGEX, 
		height*IMAGEY);

  for (tmpx = x; tmpx < width + x; tmpx++)
    for (tmpy = y; tmpy < height + y; tmpy++)
      switch (world [tmpx][tmpy]) {
      case FRIENDLY_ROBOT:
      case ENEMY_ROBOT:
	getRobotInfo (tmpx, tmpy, &team_id, &robot_id);
	drawRobot (current_sector, tmpx, tmpy, tmpx, tmpy, 
		   robot_front[team_id][robot_id], 
		   world[tmpx][tmpy], NULL);
	break;

      case POS_OBS:
	drawPosObs (current_sector, tmpx, tmpy);
	break;
	
      case NEG_OBS:
	drawNegObs (current_sector, tmpx, tmpy);
	break;

      case STATIC_OBS:
	drawStaticObs (current_sector, tmpx, tmpy);
	break;

      case UNKNOWN:
	drawUnknown (current_sector, tmpx, tmpy);
	break;

      case N_MISSILE:
      case E_MISSILE:
      case S_MISSILE:
      case W_MISSILE:
	break;

      default:
	coord.x = tmpx;
	coord.y = tmpy;
	if (equal (coord, goal))
	  redrawGoal (current_sector, goal);
	break;
      }

  XFlush (Disp);
}


int activeSector (char *sector, int x, int y) {

  if ((x < MAXX) && (y < MAXY) &&
      (strcmp (sector, "Top Left") == 0))
    return TRUE;

  if ((x >= MAXX) && (y < MAXY) &&
      (strcmp (sector, "Top Right") == 0))
    return TRUE;

  if ((x < MAXX) && (y >= MAXY) &&
      (strcmp (sector, "Bottom Left") == 0))
    return TRUE;

  if ((x >= MAXX) && (y >= MAXY) &&
      (strcmp (sector, "Bottom Right") == 0))
    return TRUE;

  return FALSE;
}


char *getActiveSector (int x, int y) {
  if ((x < MAXX*RADAR_X) && (y < MAXY*RADAR_Y))
    return ("Top Left");

  if ((x >= MAXX*RADAR_X) && (y < MAXY*RADAR_Y))
    return ("Top Right");

  if ((x < MAXX*RADAR_X) && (y >= MAXY*RADAR_Y))
    return ("Bottom Left");

  if ((x >= MAXX*RADAR_X) && (y >= MAXY*RADAR_Y))
    return ("Bottom Right");
}


void wipeObject (char *sector, int x, int y) {
  int init_x, init_y;

  /* Wipe from the radar window. */
  XFillRectangle (Disp, RWin, GcB, x*RADAR_X, 
		  y*RADAR_Y, RADAR_X, RADAR_Y);

  if (!activeSector (sector, x, y))
    return;

  /* Avoid robot trails being left in hilighted sector window. */
  XFillRectangle (Disp, RWin, GcSector, x*RADAR_X, 
		  y*RADAR_Y, RADAR_X, RADAR_Y);

  /* Wipe from the field window if active. */
  getSectorOffsets (sector, &init_x, &init_y);
  drawBackdrop (sector, (x-init_x)*IMAGEX, 
		(y-init_y)*IMAGEY, IMAGEX, IMAGEY);
  XFlush (Disp);
}


void drawObject (char *sector, 
		 int x, 
		 int y, 
		 int orientation, 
		 int object_type, 
		 char *data_buf) {

  /* "data_buf" is currently only used with drawRobot(). 
   It is used to specify any optional data required to draw object. */

  switch (object_type) {
  case FRIENDLY_ROBOT:
  case ENEMY_ROBOT:
    drawRobot (sector, x, y, x, y, orientation, object_type, data_buf);
    break;

  case N_MISSILE:
  case E_MISSILE:
  case S_MISSILE:
  case W_MISSILE:
    drawMissile (sector, x, y, x, y, orientation);
    break;

  case POS_OBS:
    drawPosObs (sector, x, y);
    break;

  case NEG_OBS:
    drawNegObs (sector, x, y);
    break;

  case STATIC_OBS:
    drawStaticObs (sector, x, y);
    break;

  case UNKNOWN:
    drawUnknown (sector, x, y);
    break;

  default:
    wipeObject (sector, x, y);
    break;
  }
}


void drawRadar (char *sector, int type, int from_x, 
		int from_y, int to_x, int to_y) {

  wipeObject (sector, from_x, from_y);
  
  if (type == NEG_OBS) 
    XDrawRectangle (Disp, RWin, GcNF, to_x*RADAR_X, 
		    to_y*RADAR_Y, RADAR_X-1, RADAR_Y-1);

  else if (type == STATIC_OBS) 
    XDrawRectangle (Disp, RWin, GcDF, to_x*RADAR_X, 
		    to_y*RADAR_Y, RADAR_X-1, RADAR_Y-1);
  else if (type == FRIENDLY_ROBOT)
    XFillRectangle (Disp, RWin, GcGRF, to_x*RADAR_X, 
		    to_y*RADAR_Y, RADAR_X, RADAR_Y);
  else if (type == ENEMY_ROBOT || type == UNKNOWN)
    XFillRectangle (Disp, RWin, GcBRF, to_x*RADAR_X, 
		    to_y*RADAR_Y, RADAR_X, RADAR_Y);

  else /* Goal. */
    XFillRectangle (Disp, RWin, GcGF, to_x*RADAR_X, 
		    to_y*RADAR_Y, RADAR_X, RADAR_Y);
}


void drawRobot (char *sector, 
		int from_x, 
		int from_y, 
		int to_x, 
		int to_y, 
		int orientation, 
		int type,
		char *data_buf) {

  int i, j, sign_x = 0, sign_y = 0;
  int init_x, init_y;
  int team_id, robot_id;
  Pixmap enemy_robot, pixmap;

  /* We need to be able to reproduce the enemy robot images. We could 
     use our local Pixmaps for enemy robots to do this (ie brobotn,
     brobote etc) but instead we get the pixmap image data from the
     messages received across the network. We can then build a pixmap
     from this data. This looks contrived (AND IT IS!) but it is meant 
     to simulate the burden of having to fetch image data from across
     the network. This could be valid in a real situation, when we don't
     know what an image looks like until we see it. 

     Note that only enemy robot images are passed around. All other
     object-types can be drawn locally. There were two scenarios to 
     consider:
     1. Client only knows what its team of robots looks like.
     2. Client knows what all objects look like except enemy robots.

     Option 2 was chosen because robots are the dynamic entities in this
     application. They can walk, rotate and fire missiles. For this reason,
     we could envision images about robots changing as the robots execute
     different events. All other objects are essentially static, so once
     image information is captured about these objects, it doesn't
     really change. 

     Finally: Each client still has the power to draw enemy robots without
     fetching the robot image. This is necessary in refreshWorld(), where
     the clients view of the battlefield window may need redrawing after 
     being raised from behind another window, or when a pulldown menu has
     just disappeared.
 */

  drawRadar (sector, type, from_x, from_y, to_x, to_y);

  if (!activeSector (sector, from_x, from_y)) {
    if (!activeSector (sector, to_x, to_y))
      return;
    else {
      drawRobot (sector, to_x, to_y, to_x, to_y, 
		 orientation, type, data_buf);
      return;
    }
  }

  getSectorOffsets (sector, &init_x, &init_y);

  /* The use of sign variables and the loop is a 
     bit of overkill to smooth the animation when a robot moves.
     Instead of erasing and redrawing a robot in steps of IMAGEX 
     (or IMAGEY), the robots are quickly erased and
     redrawn on a pixel-by-pixel basis. The refresh rate is 
     controlled via a Tcl/Tk animation delay variable. */

  if (to_x < from_x)
    sign_x = -1;
  else
    if (to_x > from_x)
      sign_x = +1;

  if (to_y < from_y)
    sign_y = -1;
  else
    if (to_y > from_y)
      sign_y = +1;

  if ((type == ENEMY_ROBOT) && (data_buf != NULL))
    XpmCreatePixmapFromData (Disp, Win, (char **)data_buf,
			     &enemy_robot, NULL, NULL);

  i = j = 1;

  do {
    drawBackdrop (sector, 
		  (from_x-init_x)*IMAGEX+(i-1)*sign_x, 
		  (from_y-init_y)*IMAGEY+(j-1)*sign_y, 
		  IMAGEX, 
		  IMAGEY);

    XSetClipOrigin (Disp, 
		    GcRobot[orientation],
		    (from_x-init_x)*IMAGEX+i*sign_x, 
		    (from_y-init_y)*IMAGEY+j*sign_y); 

    if (type == FRIENDLY_ROBOT)
      pixmap = robot[orientation];
    else {
      if (data_buf != NULL)
	pixmap = enemy_robot;
      else
	pixmap = brobot[orientation];
    }

    XCopyArea (Disp, 
	       pixmap, 
	       FWin, 
	       GcRobot[orientation], 
	       0, 
	       0, 
	       IMAGEX, 
	       IMAGEY, 
	       (from_x-init_x)*IMAGEX+i*sign_x, 
	       (from_y-init_y)*IMAGEY+j*sign_y);
    
    XFlush (Disp);

    i++;
    j++;
  }
  while ((i <= IMAGEX && j <= IMAGEY) && (sign_x !=0 || sign_y != 0));

  /* Add remaining life points to robot image. */
  getRobotInfo (to_x, to_y, &team_id, &robot_id);
  drawLifePoints (sector, 
		  to_x, 
		  to_y, 
		  team_id, 
		  robot_id,
		  orientation);

  if ((type == ENEMY_ROBOT) && (data_buf != NULL))
    XFreePixmap (Disp, enemy_robot);
}


void drawNegObs (char *sector, int x, int y) {
  int init_x, init_y;

  drawRadar (sector, NEG_OBS, x, y, x, y);

  if (!activeSector (sector, x, y))
    return;

  getSectorOffsets (sector, &init_x, &init_y);

  XSetClipOrigin (Disp, 
		  GcBomb, 
		  (x-init_x)*IMAGEX, 
		  (y-init_y)*IMAGEY);

  XCopyArea (Disp, 
	     bomb, 
	     FWin, 
	     GcBomb, 
	     0, 
	     0, 
	     IMAGEX, 
	     IMAGEY, 
	     (x-init_x)*IMAGEX, 
	     (y-init_y)*IMAGEY);
}


void drawPosObs (char *sector, int x, int y) {
  int init_x, init_y;

  if (!activeSector (sector, x, y))
    return;

  getSectorOffsets (sector, &init_x, &init_y);

  if ((strcmp (sector, "Top Left") == 0) ||
      (strcmp (sector, "Top Right") == 0)) {

    XSetClipOrigin (Disp, 
		    GcBonus, 
		    (x-init_x)*IMAGEX, 
		    (y-init_y)*IMAGEY);

    XCopyArea (Disp, 
	       bonus, 
	       FWin, 
	       GcBonus, 
	       0, 
	       0, 
	       IMAGEX, 
	       IMAGEY, 
	       (x-init_x)*IMAGEX, 
	       (y-init_y)*IMAGEY);
  }
  else {
    XSetClipOrigin (Disp, 
		    GcBonus, 
		    (x-init_x)*IMAGEX, 
		    (y-init_y)*IMAGEY);

    XCopyArea (Disp, 
	       bonus, 
	       FWin, 
	       GcBonus, 
	       0, 
	       0, 
	       IMAGEX, 
	       IMAGEY, 
	       (x-init_x)*IMAGEX, 
	       (y-init_y)*IMAGEY);
  }
}


void drawUnknown (char *sector, int x, int y) {
 int init_x, init_y;

  drawRadar (sector, UNKNOWN, x, y, x, y);

  if (!activeSector (sector, x, y))
    return;

  getSectorOffsets (sector, &init_x, &init_y);

  XSetClipOrigin (Disp, 
		  GcUnknown, 
		  (x-init_x)*IMAGEX, 
		  (y-init_y)*IMAGEY);

  XCopyArea (Disp, 
	     unknown, 
	     FWin, 
	     GcUnknown, 
	     0, 
	     0, 
	     IMAGEX, 
	     IMAGEY, 
	     (x-init_x)*IMAGEX, 
	     (y-init_y)*IMAGEY);
}


void drawStaticObs (char *sector, int x, int y) {
  int init_x, init_y;

  drawRadar (sector, STATIC_OBS, x, y, x, y);

  if (!activeSector (sector, x, y))
    return;

  getSectorOffsets (sector, &init_x, &init_y);

  if ((strcmp (sector, "Top Left") == 0) ||
      (strcmp (sector, "Bottom Left") == 0)) {

    XSetClipOrigin (Disp, 
		    GcTree, 
		    (x-init_x)*IMAGEX, 
		    (y-init_y)*IMAGEY);

    XCopyArea (Disp, 
	       tree, 
	       FWin, 
	       GcTree, 
	       0, 
	       0, 
	       IMAGEX, 
	       IMAGEY, 
	       (x-init_x)*IMAGEX, 
	       (y-init_y)*IMAGEY);
  }
  else {
    XSetClipOrigin (Disp, 
		    GcBlock, 
		    (x-init_x)*IMAGEX, 
		    (y-init_y)*IMAGEY);

    XCopyArea (Disp, 
	       block, 
	       FWin, 
	       GcBlock, 
	       0, 
	       0, 
	       IMAGEX, 
	       IMAGEY, 
	       (x-init_x)*IMAGEX, 
	       (y-init_y)*IMAGEY);
  }
}


void drawMissile (char *sector, int from_x, int from_y, 
		  int to_x, int to_y, int orientation) {
  int i, j, sign_x = 0, sign_y = 0;
  int init_x, init_y;

  if (!activeSector (sector, from_x, from_y))
    return;

  getSectorOffsets (sector, &init_x, &init_y);

  if (to_x < from_x)
    sign_x = -1;
  else
    if (to_x > from_x)
      sign_x = +1;

  if (to_y < from_y)
    sign_y = -1;
  else
    if (to_y > from_y)
      sign_y = +1;

  for (i = 1, j = 1; i <= IMAGEX, j <= IMAGEY; i++, j++) {

    drawBackdrop (sector, 
		  (from_x-init_x)*IMAGEX+(i-1)*sign_x, 
		  (from_y-init_y)*IMAGEY+(j-1)*sign_y, 
		  IMAGEX, 
		  IMAGEY);

    XSetClipOrigin (Disp, 
		    GcRocket[orientation], 
		    (from_x-init_x)*IMAGEX+i*sign_x, 
		    (from_y-init_y)*IMAGEY+j*sign_y);

    XCopyArea (Disp, 
	       rocket[orientation], 
	       FWin, 
	       GcRocket[orientation], 
	       0, 
	       0, 
	       IMAGEX, 
	       IMAGEY, 
	       (from_x-init_x)*IMAGEX+i*sign_x, 
	       (from_y-init_y)*IMAGEY+j*sign_y);

    XFlush (Disp);
  }
}


void doExplosion (char *sector, int x, int y) {
  int init_x, init_y;
  int i;

  XFillArc (Disp, 
	    RWin, 
	    GcEF, 
	    (int) x*RADAR_X-RADAR_X, 
	    (int) y*RADAR_Y-RADAR_Y, 
 	    RADAR_X*3, 
	    RADAR_Y*3, 
	    0, 
	    360*64);

  XFlush (Disp); 
  snooze (50000);

  XFillArc (Disp, 
	    RWin, 
	    GcEF, 
	    (int) x*RADAR_X-RADAR_X,
 	    (int) y*RADAR_Y-RADAR_Y, 
 	    RADAR_X*3, 
	    RADAR_Y*3, 
	    0, 
	    360*64);

  XFlush (Disp);

  if (!activeSector (sector, x, y))
    return;

  getSectorOffsets (sector, &init_x, &init_y);

  for (i = 0; i < EXP_FRAMES; i++) {

    XSetClipOrigin (Disp, 
		    GcExp[i], 
		    (x-init_x)*IMAGEX-20, 
		    (y-init_y)*IMAGEY-20); 

    XCopyArea (Disp, 
	       Exp[i], 
	       FWin, 
	       GcExp[i], 
	       0, 
	       0, 
	       80, 
	       80, 
	       (x-init_x)*IMAGEX-20, 
	       (y-init_y)*IMAGEY-20);
    XFlush (Disp);
    snooze (50000);
  }

  /* Redraw objects affected by explosion. */
  refreshWorld (current_sector, 
		sector, 
		x-1, 
		y-1, 
		3, 
		3);

  XFlush (Disp);
}


void glorySequence (char *sector, int x, int y, int type) {
  int i, j;
  int init_x, init_y;

  if (!activeSector (sector, x, y)) {
    XFlush (Disp);
    return;
  }

  getSectorOffsets (sector, &init_x, &init_y);

  XFillArc (Disp, 
	    FWin, 
	    GcEF, 
	    (int) (x-init_x)*IMAGEX-IMAGEX, 
	    (int) (y-init_y)*IMAGEY-IMAGEY, 
	    IMAGEX*3, 
	    IMAGEY*3, 
	    0, 
	    360*64);

  XFillArc (Disp, 
	    FWin, 
	    GcEF, 
	    (int) (x-init_x)*IMAGEX-IMAGEX/2, 
	    (int) (y-init_y)*IMAGEY-IMAGEY/2, 
	    IMAGEX*2, 
	    IMAGEY*2, 
	    0, 
	    360*64);

  for (i = 0; i < 10; i++) {

    if (type == FRIENDLY_ROBOT) {
      for (j = 0; j < FULL_ROTATION; j++) {
	XSetClipOrigin (Disp, 
			GcRobot[j], 
			(x-init_x)*IMAGEX, 
			(y-init_y)*IMAGEY);
	
	XCopyArea (Disp, 
		   robot[j], 
		   FWin, 
		   GcRobot[j], 
		   0, 
		   0, 
		   IMAGEX, 
		   IMAGEY, 
		   (x-init_x)*IMAGEX, 
		   (y-init_y)*IMAGEY);
	
	XFlush (Disp);
	snooze ((int) (robot_delay*1000)/MISSILE_VELOCITY);
      }
    }
    else {
      for (j = 0; j < FULL_ROTATION; j++) {
	XSetClipOrigin (Disp, 
			GcRobot[j], 
			(x-init_x)*IMAGEX, 
			(y-init_y)*IMAGEY);
	
	XCopyArea (Disp, 
		   brobot[j], 
		   FWin, 
		   GcRobot[j], 
		   0, 
		   0, 
		   IMAGEX, 
		   IMAGEY, 
		   (x-init_x)*IMAGEX, 
		   (y-init_y)*IMAGEY);
	
	XFlush (Disp);
	snooze ((int) (robot_delay*1000)/MISSILE_VELOCITY);
      }
    }
  }
  
  XFillArc (Disp, 
	    FWin, 
	    GcEF, 
	    (int) 
	    (x-init_x)*IMAGEX-IMAGEX/2, 
	    (int) (y-init_y)*IMAGEY-IMAGEY/2, 
	    IMAGEX*2, 
	    IMAGEY*2, 
	    0, 
	    360*64);
  
  XFillArc (Disp,  
	    FWin, 
	    GcEF, 
	    (int) (x-init_x)*IMAGEX-IMAGEX, 
	    (int) (y-init_y)*IMAGEY-IMAGEY, 
	    IMAGEX*3, 
	    IMAGEY*3, 
	    0, 
	    360*64);

  XFlush (Disp);
}


void redrawGoal (char *sector, coordinate goal) {
  int init_x, init_y;

  drawRadar (sector, -1, goal.x, goal.y, goal.x, goal.y);

  if (!activeSector (sector, goal.x, goal.y))
    return;

  getSectorOffsets (sector, &init_x, &init_y);

  XCopyArea (Disp, 
	     target, 
	     FWin, 
	     GcGF, 
	     0, 
	     0, 
	     IMAGEX, 
	     IMAGEY, 
	     (goal.x-init_x)*IMAGEX, 
	     (goal.y-init_y)*IMAGEY);

  XFlush (Disp);
}


void maskGraphicsExposures () {
  int i;

  XSetGraphicsExposures (Disp, GcDF, FALSE);
  XSetGraphicsExposures (Disp, GcEF, FALSE);
  XSetGraphicsExposures (Disp, GcBomb, FALSE);
  XSetGraphicsExposures (Disp, GcBonus, FALSE);
  XSetGraphicsExposures (Disp, GcTree, FALSE);
  XSetGraphicsExposures (Disp, GcBlock, FALSE);
  XSetGraphicsExposures (Disp, GcUnknown, FALSE);
  XSetGraphicsExposures (Disp, GcNum, FALSE);

  for (i = 0; i < FULL_ROTATION; i++) {
    XSetGraphicsExposures (Disp, GcRocket[i], FALSE);
    XSetGraphicsExposures (Disp, GcRobot[i], FALSE);
  }
  for (i = 0; i < EXP_FRAMES; i++)
    XSetGraphicsExposures (Disp, GcExp[i], FALSE);
}


void Xinit (int argc, char *argv[]) {      /* Initialise X. */
  char *window_name = "Battle Robots";
  char *icon_name = "Battle Robots";
  Pixmap icon_pixmap;
  XGCValues gcv;
  char *DisplayName = NULL;
  XWMHints wm_hints;
  XFontStruct *font_info;
  char *font1 = "-adobe-helvetica-bold-r-normal-*-12-*-75-75-*-*-iso8859-1";
  char *font2 = "-adobe-helvetica-bold-r-*-*-12-*-*-*-*-*-*-*";
  XTextProperty windowName, iconName;
  unsigned long bg,fg;
  Cursor Curs;
  int i;

  DisplayName = (char *)getenv("DISPLAY");
  if(DisplayName==NULL)
    {
      fprintf(stderr,"%s: $DISPLAY not set\n",progname);
      shutdown (1);
    }
  
  Disp=XOpenDisplay (DisplayName);

  if(Disp==NULL)
    {
      fprintf(stderr,"%s: Unable to open $DISPLAY (%s)\n",
	      progname,DisplayName);
      shutdown (1);
    }

  Root = XDefaultRootWindow (Disp);
  Scr = DefaultScreen (Disp);
  colormap = DefaultColormap (Disp, Scr);
  initColors ();

  fg = BlackPixel (Disp,Scr);
  if ((bg = colorPixelValues[2]) == fg)     /* black and white */
    bg = WhitePixel (Disp, Scr);

  /* Main (parent) window */
  Win = Tk_WindowId (Et_MainWindow); 
  /* Have now created the main window from Tcl/Tk using Et_Init. */

  /* Field window */
  FWin = XCreateSimpleWindow (Disp,Win,BORDER,BORDER+MENU_HEIGHT,
			      MAXX*IMAGEX,MAXY*IMAGEY,1,fg,bg);

  RWin = XCreateSimpleWindow (Disp,Win,4*BORDER+MAXX*IMAGEX,
			      BORDER+MENU_HEIGHT+MAXY*IMAGEY-RADAR_HEIGHT,
			      RADAR_WIDTH,RADAR_HEIGHT,
			      1,fg,bg);

  /* Control window 1 - START */
  CWin1 = XCreateSimpleWindow (Disp,Win,4*BORDER+MAXX*IMAGEX,
			       5*BORDER+STATUS_HEIGHT+
			       CONTROL_HEIGHT+MENU_HEIGHT,
			       CONTROL_WIDTH,CONTROL_HEIGHT,1,fg,
			       colorPixelValues[8]);

  /* Control window 2 - STOP */
  CWin2 = XCreateSimpleWindow (Disp,Win,6*BORDER+MAXX*IMAGEX+CONTROL_WIDTH,
			       5*BORDER+STATUS_HEIGHT+CONTROL_HEIGHT+
			       MENU_HEIGHT,
			       CONTROL_WIDTH,CONTROL_HEIGHT,1,fg,
			       colorPixelValues[4]);

  /* Control window 3 - PAUSE */
  CWin3 = XCreateSimpleWindow (Disp,Win,4*BORDER+MAXX*IMAGEX,
			       3*BORDER+STATUS_HEIGHT+MENU_HEIGHT,
			       CONTROL_WIDTH*2+2*BORDER,CONTROL_HEIGHT,1,fg,
			       colorPixelValues[10]);

  gcv.function = GXcopy;
  gcv.plane_mask = AllPlanes;
  gcv.fill_style = FillTiled;

  for (i = 0; i < NUM_SECTORS; i++)
    GcBattle[i] = XCreateGC (Disp,
			     Win,
			     (GCFunction | GCPlaneMask | GCFillStyle),
			     &gcv);

  gcv.fill_style = FillSolid;

  GcDF=XCreateGC (Disp,Win,(GCFunction | GCPlaneMask | GCFillStyle),&gcv);
  XSetForeground (Disp,GcDF,fg);

  GcB=XCreateGC (Disp,Win,(GCFunction | GCPlaneMask | GCFillStyle),&gcv);
  XSetForeground (Disp,GcB,bg);
  
  GcPF=XCreateGC (Disp,Win,(GCFunction | GCPlaneMask | GCFillStyle),&gcv);
  XSetForeground (Disp,GcPF,colorPixelValues[11]);

  GcBRF=XCreateGC (Disp,Win,(GCFunction | GCPlaneMask | GCFillStyle),&gcv);
  XSetForeground (Disp,GcBRF,colorPixelValues[7]);

  GcGRF=XCreateGC (Disp,Win,(GCFunction | GCPlaneMask | GCFillStyle),&gcv);
  XSetForeground (Disp,GcGRF,colorPixelValues[4]);

  GcNF=XCreateGC (Disp,Win,(GCFunction | GCPlaneMask | GCFillStyle),&gcv);
  XSetForeground (Disp,GcNF,colorPixelValues[0]);

  GcGF=XCreateGC (Disp,Win,(GCFunction | GCPlaneMask | GCFillStyle),&gcv);
  XSetForeground (Disp,GcGF,colorPixelValues[5]);

  /* These Graphic Contexts are specific to the pixmaps on the battlefield. */
  GcBomb = XCreateGC (Disp,Win,(GCFunction | GCPlaneMask | GCFillStyle), 
		      &gcv);

  GcBonus = XCreateGC (Disp,Win,(GCFunction | GCPlaneMask | GCFillStyle),
		       &gcv);

  GcTree = XCreateGC (Disp,Win,(GCFunction | GCPlaneMask | GCFillStyle),
		      &gcv);

  GcBlock = XCreateGC (Disp,Win,(GCFunction | GCPlaneMask | GCFillStyle),
		       &gcv);

  GcUnknown = XCreateGC (Disp,Win,(GCFunction | GCPlaneMask | GCFillStyle),
			 &gcv);

  GcNum = XCreateGC (Disp,Win,(GCFunction | GCPlaneMask | GCFillStyle),
		     &gcv);

  for (i = 0; i < FULL_ROTATION; i++) {
    GcRobot[i] = XCreateGC (Disp,Win,(GCFunction | GCPlaneMask | GCFillStyle),
			    &gcv);
    GcRocket[i] = XCreateGC (Disp,Win,(GCFunction | GCPlaneMask | GCFillStyle),
			     &gcv);
  }

  for (i = 0; i < EXP_FRAMES; i++)
    GcExp[i] = XCreateGC (Disp,Win,(GCFunction | GCPlaneMask | GCFillStyle),
			  &gcv);

  gcv.function = GXxor;

  GcSector = XCreateGC (Disp,Win,(GCFunction | GCPlaneMask | GCFillStyle),
			&gcv);

  XSetForeground (Disp,GcSector,colorPixelValues[9] ^ bg);
  XSetBackground (Disp,GcSector,bg);

  GcTF = XCreateGC (Disp,Win,(GCFunction | GCPlaneMask | GCFillStyle),
		    &gcv);

  XSetForeground (Disp,GcTF,colorPixelValues[0] ^ colorPixelValues[6]);
  XSetBackground (Disp,GcTF,colorPixelValues[6]);

  GcEF = XCreateGC (Disp,Win,(GCFunction | GCPlaneMask | GCFillStyle),
		    &gcv);

  XSetForeground (Disp,GcEF,colorPixelValues[3] ^ bg);
  XSetBackground (Disp,GcEF,bg);

  Curs = XCreateFontCursor (Disp,XC_diamond_cross);
  XDefineCursor (Disp,Win,Curs);

  XpmCreatePixmapFromData (Disp,
			   Win,
			   icon_pixmap_xpm,
			   &icon_pixmap,
			   NULL,
			   NULL);

  wm_hints.flags = (InputHint|IconPixmapHint);
  wm_hints.input = True;
  wm_hints.icon_pixmap = icon_pixmap;

  if (argc < 3) { 
    /* No name for main window or icon given, so create defaults. */
    if (XStringListToTextProperty (&window_name, 1, &windowName) == 0) {
      (void) fprintf (stderr, 
		      "%s: structure allocation for windowName failed.\n", 
		      progname);
      shutdown(-1);
    }

    if (XStringListToTextProperty (&icon_name, 1, &iconName) == 0) {
      (void) fprintf (stderr, 
		      "%s: structure allocation for iconName failed.\n", 
		      progname);
      shutdown(-1);
    }
    XSetWMProperties (Disp, Win, &windowName, &iconName, 
		      argv, argc, NULL, &wm_hints, NULL);
  }
  else
    XSetWMProperties (Disp, Win, NULL, NULL, 
		      argv, argc, NULL, &wm_hints, NULL);

  if ((font_info = XLoadQueryFont(Disp, font1)) == NULL) {
    fprintf( stderr, "Battle Robots: Cannot open %s font\n", font1);
    shutdown (1);
  }
  XSetFont (Disp, GcDF, font_info->fid);
  XSetFont (Disp, GcTF, font_info->fid);

  if ((font_info = XLoadQueryFont(Disp, font2)) == NULL) {
    fprintf( stderr, "Battle Robots: Cannot open %s font\n", font2);
    shutdown (1);
  }
  XSetFont (Disp, GcGF, font_info->fid);

  maskGraphicsExposures ();
}


void handleXpmError (int error_status) {
  char *warning = NULL;
  char *error   = NULL;

  /* Switch on the type of error returned by xpm library */
  switch (error_status) {
  case XpmSuccess:
    return;

  case XpmColorError:
    warning = "Could not parse or alloc requested colour";
    break;

  case XpmNoMemory:
    error = "Not enough memory for pixmap creation";
    break;

  case XpmColorFailed:
    error = "Colourmap is full - cannot allocate a colour";
    break;

  case XpmOpenFailed:
    error = "Unable to open pixmap file";
    break;

  case XpmFileInvalid:
    error = "XPM file contains invalid or corrupt data";
    break;

  default:
    error = "Unexpected xpm error code";
    break;
  }

  if (warning)
    fprintf (stderr, "(Warning) %s.\n", warning);
  else if (error) {
    fprintf (stderr, "(Error) %s.\n", error);
    shutdown (1);
  }
  else
    return;
}


void setupPixmaps (void) {
  int i;

  handleXpmError (XpmCreatePixmapFromData 
		  (Disp,FWin,goal_xpm, 
		   &target,NULL,NULL));

  handleXpmError (XpmCreatePixmapFromData 
		  (Disp,FWin,bomb_xpm, 
		   &bomb,&bomb_mask,NULL));

  handleXpmError (XpmCreatePixmapFromData 
		  (Disp,FWin,bonus_xpm, 
		   &bonus,&bonus_mask,NULL));

  handleXpmError (XpmCreatePixmapFromData 
		  (Disp,FWin,tree_xpm, 
		   &tree,&tree_mask,NULL));

  handleXpmError (XpmCreatePixmapFromData 
		  (Disp,FWin,block_xpm, 
		   &block,&block_mask,NULL));

  handleXpmError (XpmCreatePixmapFromData 
		  (Disp,FWin,unknown_xpm, 
		   &unknown,&unknown_mask,NULL));

  handleXpmError (XpmCreatePixmapFromData 
		  (Disp,FWin,rocketn_xpm, 
		   &rocket[0],&rocket_mask[0],NULL));

  handleXpmError (XpmCreatePixmapFromData 
		  (Disp,FWin,rockete_xpm, 
		   &rocket[1],&rocket_mask[1],NULL));

  handleXpmError (XpmCreatePixmapFromData 
		  (Disp,FWin,rockets_xpm, 
		   &rocket[2],&rocket_mask[2],NULL));

  handleXpmError (XpmCreatePixmapFromData 
		  (Disp,FWin,rocketw_xpm, 
		   &rocket[3],&rocket_mask[3],NULL));

  handleXpmError (XpmCreatePixmapFromData 
		  (Disp,FWin,robotn_xpm, 
		   &robot[0],&robot_mask[0],NULL));

  handleXpmError (XpmCreatePixmapFromData 
		  (Disp,FWin,robote_xpm, 
		   &robot[1],&robot_mask[1],NULL));

  handleXpmError (XpmCreatePixmapFromData 
		  (Disp,FWin,robots_xpm, 
		   &robot[2],&robot_mask[2],NULL));

  handleXpmError (XpmCreatePixmapFromData 
		  (Disp,FWin,robotw_xpm, 
		   &robot[3],&robot_mask[3],NULL));

  handleXpmError (XpmCreatePixmapFromData 
		  (Disp,FWin,brobotn_xpm, 
		   &brobot[0],NULL,NULL));

  handleXpmError (XpmCreatePixmapFromData 
		  (Disp,FWin,brobote_xpm, 
		   &brobot[1],NULL,NULL));

  handleXpmError (XpmCreatePixmapFromData 
		  (Disp,FWin,brobots_xpm, 
		   &brobot[2],NULL,NULL));

  handleXpmError (XpmCreatePixmapFromData 
		  (Disp,FWin,brobotw_xpm, 
		   &brobot[3],NULL,NULL));

  handleXpmError (XpmCreatePixmapFromData 
		  (Disp,FWin,explosion_0_xpm, 
		   &Exp[0],&exp_mask[0],NULL));

  handleXpmError (XpmCreatePixmapFromData 
		  (Disp,FWin,explosion_1_xpm, 
		   &Exp[1],&exp_mask[1],NULL));

  handleXpmError (XpmCreatePixmapFromData 
		  (Disp,FWin,explosion_2_xpm, 
		   &Exp[2],&exp_mask[2],NULL));

  handleXpmError (XpmCreatePixmapFromData 
		  (Disp,FWin,explosion_3_xpm, 
		   &Exp[3],&exp_mask[3],NULL));

  handleXpmError (XpmCreatePixmapFromData 
		  (Disp,FWin,explosion_4_xpm, 
		   &Exp[4],&exp_mask[4],NULL));

  handleXpmError (XpmCreatePixmapFromData 
		  (Disp,FWin,explosion_5_xpm, 
		   &Exp[5],&exp_mask[5],NULL));

  handleXpmError (XpmCreatePixmapFromData 
		  (Disp,FWin,bg_green_xpm, 
		   &bg[0],NULL,NULL));

  handleXpmError (XpmCreatePixmapFromData 
		  (Disp,FWin,bg_desert_xpm, 
		   &bg[1],NULL,NULL));

  handleXpmError (XpmCreatePixmapFromData 
		  (Disp,FWin,bg_gray_xpm, 
		   &bg[2],NULL,NULL));

  handleXpmError (XpmCreatePixmapFromData 
		  (Disp,FWin,bg_red_xpm, 
		   &bg[3],NULL,NULL));

  handleXpmError (XpmCreatePixmapFromData 
		  (Disp,FWin,one_xpm, 
		   &num[0],&num_mask,NULL));

  handleXpmError (XpmCreatePixmapFromData 
		  (Disp,FWin,two_xpm, 
		   &num[1],&num_mask,NULL));

  handleXpmError (XpmCreatePixmapFromData 
		  (Disp,FWin,three_xpm, 
		   &num[2],&num_mask,NULL));

  handleXpmError (XpmCreatePixmapFromData 
		  (Disp,FWin,four_xpm, 
		   &num[3],&num_mask,NULL));

  handleXpmError (XpmCreatePixmapFromData 
		  (Disp,FWin,five_xpm, 
		   &num[4],&num_mask,NULL));

  handleXpmError (XpmCreatePixmapFromData 
		  (Disp,FWin,six_xpm, 
		   &num[5],&num_mask,NULL));

  handleXpmError (XpmCreatePixmapFromData 
		  (Disp,FWin,seven_xpm, 
		   &num[6],&num_mask,NULL));

  handleXpmError (XpmCreatePixmapFromData 
		  (Disp,FWin,eight_xpm, 
		   &num[7],&num_mask,NULL));

  handleXpmError (XpmCreatePixmapFromData 
		  (Disp,FWin,nine_xpm, 
		   &num[8],&num_mask,NULL));

  handleXpmError (XpmCreatePixmapFromData 
		  (Disp,FWin,ten_xpm, 
		   &num[9],&num_mask,NULL));

  for (i = 0; i < NUM_SECTORS; i++)
    XSetTile (Disp, GcBattle[i], bg[i]);

  XSetClipMask (Disp, GcBomb, bomb_mask);
  XSetClipMask (Disp, GcBonus, bonus_mask);
  XSetClipMask (Disp, GcTree, tree_mask);
  XSetClipMask (Disp, GcBlock, block_mask);
  XSetClipMask (Disp, GcUnknown, unknown_mask);
  XSetClipMask (Disp, GcNum, num_mask);

  for (i = 0; i < FULL_ROTATION; i++) {
    XSetClipMask (Disp, GcRobot[i], robot_mask[i]);
    XSetClipMask (Disp, GcRocket[i], rocket_mask[i]);
  }
  
  for (i = 0; i < EXP_FRAMES; i++)
    XSetClipMask (Disp, GcExp[i], exp_mask[i]);
}
