#include "consistency.h"

extern unsigned long current_time;
extern client_t client [NUM_TEAMS];
extern void sendMsg (int conn_id, void *msg, 
		     int msg_size, char *error_msg);
extern int socketHandle (int timeout);

static textmsg_t stext;
static writers_t *w;

/* Pass pointer to a dlock_t structure, containing the manager
 for the corresponding lock. Return TRUE if this client had to
 send a request to a remote host, or FALSE if it sent to itself.*/

void sendAcqLock (int cid, dlock_t *l) {

  /* Request lock from appropriate manager.  */
  if (l->mgr != id) {
    memset (&stext, -1, sizeof (stext));
    stext.m                    = OUTSWAP (text);
    stext.id                   = OUTSWAP (cid);
    stext.type                 = OUTSWAP (MGR_LK_ACQ);
    stext.info.dlock.lock_id   = OUTSWAP (l->lock_id);
    stext.info.dlock.type      = OUTSWAP (l->type);
    stext.info.dlock.ts        = OUTSWAP (l->ts);
    stext.info.dlock.client_id = OUTSWAP (l->client_id);
    
    sendMsg (client[l->mgr].conn_id, 
	     &stext,
	     sizeof (stext),
	     "sendAcqLock");
  }
  else
    acquireLockHandler (cid, TRUE, l);
}


void sendRelLock (int cid, dlock_t *l) {

  /* Send release of lock to appropriate manager.  */

  if (l->modified)
    l->ts++;

  if (l->mgr != id) {
    memset (&stext, -1, sizeof (stext));
    stext.m                    = OUTSWAP (text);
    stext.id                   = OUTSWAP (cid);
    stext.type                 = OUTSWAP (MGR_LK_REL);
    stext.info.dlock.lock_id   = OUTSWAP (l->lock_id);
    stext.info.dlock.ts        = OUTSWAP (l->ts);
    stext.info.dlock.modified  = OUTSWAP (l->modified);
    stext.info.dlock.client_id = OUTSWAP (l->client_id);
    stext.info.dlock.type      = OUTSWAP (l->type);

    sendMsg (client[l->mgr].conn_id, 
	     &stext,
	     sizeof (stext),
	     "sendRelLock");
  }
  else
    releaseLockHandler (cid, l);

  l->status = unlocked; /* Failsafe: Always ensure this is reset. */
}


int acquireLock (int cid,
		 dlock_t *l, 
		 int x,
		 int y) {

  static int flag = FALSE;

  if (flag == FALSE) {
    flag = TRUE; /* Only send once. */
    sendAcqLock (cid, l);
  }

  /* CLIENT_LK_ACQ is returned when client successfully
     acquires the lock. */
  if (l->mgr == id)
    if ((l->type == LK_READ) || (lock_granted)) {
      lock_granted = FALSE;
      flag = FALSE;
      return (CLIENT_LK_ACQ);
    }
  
  if (socketHandle (0) == CLIENT_LK_ACQ) {
    flag = FALSE;
    return (CLIENT_LK_ACQ);
  }
  
  return (FALSE);
}


void acquireLockHandler (int cid, int flag, dlock_t *l) {

  int x, y;
  writers_p old_pending_writers;
  int recipient; /* Client who will get the lock. */
  dlock_t dlock;

  /* Translate lock_id into an object coordinate. */
  x = lock_tt[l->lock_id].x;
  y = lock_tt[l->lock_id].y;

  recipient = cid;

#ifdef DEBUG
  printf ("(acquireLockHandler) Acquire lock for "
	  "(%d, %d), type %d, client %d, lock_granted=%d\n",
	  x, y, l->type, cid, lock_granted);
#endif

  if (l->type == LK_WRITE) {

    if (flag) {
      w = (writers_p) malloc (sizeof (writers_t));
      w->client_id = cid;
      w->ts = l->ts;
      w->next = NULL;
    
      old_pending_writers = mgr_lock[x][y].pending_writers;
    
      mgr_lock[x][y].pending_writers = 
	addToLockList (mgr_lock[x][y].pending_writers, w);

      /* If there were pending writers at the time this
	 lock request arrived simply return. */
      if (old_pending_writers) {
	/* ?? lock_granted = FALSE;*/
	return;
      }
    }
    else {
      if (l->pending_writers)
	recipient = l->pending_writers->client_id;
      else {
	/* ?? lock_granted = FALSE;*/
	return;
      }
    }
  }

  memset (&stext, -1, sizeof (stext));
  stext.m                    = OUTSWAP (text);
  stext.id                   = OUTSWAP (id);
  stext.type                 = OUTSWAP (CLIENT_LK_ACQ);
  stext.info.dlock.lock_id   = OUTSWAP (mgr_lock[x][y].lock_id);
  
  if (l->type == LK_WRITE) {
    if (mgr_lock[x][y].ts > mgr_lock[x][y].pending_writers->ts)
      stext.info.dlock.modified  = OUTSWAP (TRUE);
    else
      stext.info.dlock.modified  = 
	OUTSWAP (FALSE);
  }
  else { /* LK_READ */
    if (mgr_lock[x][y].ts > l->ts)
      stext.info.dlock.modified  = OUTSWAP (TRUE);
    else
      stext.info.dlock.modified  = OUTSWAP (FALSE);
  }
  stext.info.dlock.client_id = OUTSWAP (mgr_lock[x][y].client_id);
  
  if (recipient != id)
    sendMsg (client[recipient].conn_id, 
	     &stext,
	     sizeof (stext),
	     "acquireLockHandler");
  else {
    dlock.lock_id  = INSWAP (stext.info.dlock.lock_id);
    dlock.modified = INSWAP (stext.info.dlock.modified);
    dlock.client_id= INSWAP (stext.info.dlock.client_id);
    
    x = lock_tt[dlock.lock_id].x;
    y = lock_tt[dlock.lock_id].y;
    client_lock[x][y].modified  = dlock.modified;
    client_lock[x][y].client_id = dlock.client_id;
    /* Grant this client the lock and set the status to locked
       if it is an exclusive write lock. */
    if (client_lock[x][y].type == LK_WRITE)
      client_lock[x][y].status = locked;
    lock_granted = TRUE;
  }
}


void releaseLockHandler (int cid, dlock_t *l) {
  
  int x, y;

  /* Translate lock_id into an object coordinate. */
  x = lock_tt[l->lock_id].x;
  y = lock_tt[l->lock_id].y;

  if (l->type == LK_WRITE) {
    if (l->modified) {
      mgr_lock[x][y].ts        = l->ts;
      mgr_lock[x][y].client_id = cid;
    }
    
    delFromLockList (&mgr_lock[x][y].pending_writers);
    
    if (mgr_lock[x][y].pending_writers)
      acquireLockHandler (cid, FALSE, &mgr_lock[x][y]);
  }

#ifdef DEBUG
  printf ("(releaseLockHandler) Release lock for "
	  "(%d, %d), type %d, lock_granted=%d\n",
	  x, y, l->type, lock_granted);
#endif
}


/* Add to FIFO queue of pending writers to locked objects. */
writers_t *addToLockList (writers_t *front, 
		    writers_t *w) {

  writers_t *new_front=front;

  while (front) {
    if (front->next)
      front = front->next;
    else
      break;
  }
  
  if (front == NULL) {
    front = w;
    new_front = front;
    front->next = NULL;
  }
  else {
    front->next  = w;
    (front->next)->next = NULL;
  }

  return (new_front);
}


/* Delete from FIFO queue of pending writers to locked objects. */
void delFromLockList (writers_t **front) {

  writers_t *ptr;

  ptr = *front;
  
  if (*front != NULL) {
    *front = (*front)->next;
    free (ptr);
  }
}







