#include "scheduler.h"
#include "tk.h"
#include "socklink.h"
#include "Xdata.h"
#include "comms.h"

char *progname;
int end_prog;
int sched_num = 1;         /* Default scheduler for this client is the 
			      first scheduler in dbase.                 */
int robot_delay = 200;
int sound;
int team_size = -1;	   /* No robots yet defined.                    */

char *sector;              /* Name of currently in-view battlefield 
			      sector. A Tcl link var.                   */
char *current_sector;      /* For consistency - observe sector changes. */
char *opmode;              /* Interactive or not. A Tcl link var.
			      Not an enum type because strings link 
			      easier with Tcl.                          */

static char *server = NULL;
int sconn_id;		   /* Connection id of this client to server.   */
int id = -1;               /* ID of this client.                        */


/* Setup the Tcl/Tk links to C code. */
void Et_Setup (int *argc, char *argv[]) {

  int width, height, code;
  char geometry [10];

  Et_Init (argc, argv);

  /* Setup a main window big enough for viewable battlefield 
     window and status/control panels. Generate this
     window via Tcl/Tk and let Xinit() create
     the children using Xlib routines. */

  width = MAXX*IMAGEX+STATUS_WIDTH+6*BORDER;
  height = MAXY*IMAGEY+4*BORDER+MENU_HEIGHT;

  sprintf (geometry, "%d%c%d", width, 'x', height);

  Tcl_SetVar (Et_Interp, "geometry", geometry, TCL_GLOBAL_ONLY);
  code = Tcl_VarEval (Et_Interp, "wm geometry . ", geometry, (char *) NULL);

  if (code != TCL_OK)
    fprintf(stderr, "%s\n", Et_Interp->result);

  /* Map main application window generated by Tcl/Tk initialisation 
     before calling Xinit(). This is necessary so that main window
     exists before any subwindows are created using Xlib.
     */

  Tk_MapWindow (Et_MainWindow);

  /* The menu script. */
  ET_INCLUDE( menu1.tcl );
  
  /* Link C to any Tcl variables as necessary. */
  code = Tcl_LinkVar (Et_Interp, "anim_delay", 
		      (char *) &robot_delay, TCL_LINK_INT);
  if (code != TCL_OK)
    fprintf(stderr, "%s\n", Et_Interp->result);
  code = Tcl_LinkVar (Et_Interp, "team_size", 
		      (char *) &team_size, TCL_LINK_INT);
  if (code != TCL_OK)
    fprintf(stderr, "%s\n", Et_Interp->result);
  code = Tcl_LinkVar (Et_Interp, "end_prog", 
		      (char *) &end_prog, TCL_LINK_INT);
  if (code != TCL_OK)
    fprintf(stderr, "%s\n", Et_Interp->result);

  code = Tcl_LinkVar (Et_Interp, "sound", 
		      (char *) &sound, TCL_LINK_INT);
  if (code != TCL_OK)
    fprintf(stderr, "%s\n", Et_Interp->result);
}


void parse_args (int argc, char *argv[]) {
  int i;
  char *p, next_char;
  
  for (i = 1; i < argc; i++) {
    p = argv[i];
    if (*p++ == '-')
      switch (*p++) {

      case 'm':
	if (*p =='\0')
	  server = argv[++i];
	else
	  server = p;
	break;

      case 'n': /* Used for specifying window/icon name, 
		   but parsed by Tcl Interpreter. */
	break;

      case 's': /* Specify the schedule number. 
		   Currently EDF_SP=0 and FCFS_DD=1.
		   As schedulers are added, 
		   their constants will be added to scheduler.h. */
	if (*p =='\0')
	  sched_num = atoi (argv[++i]);
	else
	  sched_num = atoi (p);
	if ((sched_num < 0) || (sched_num > (NUM_SCHEDULERS-1)))
	  sched_num = 0;
	break;

      default:
	fprintf (stderr, "Usage: %s -m server_host"
		 "[-s sched_num] [-name \"window title\"]\n", argv[0]);
      }
  }

  if (server == NULL) {
    fprintf (stderr, "%s: No server host (-m) defined.\n", argv[0]);
    shutdown (1);
  }
}


void main (int argc, char *argv[]) {
  int args;
  int i, j;
  int port;

  parse_args (argc, argv);	/* Parse command-line args. */

  progname = argv[0];
  args = argc;
  opmode = (char *) malloc (strlen ("Computer"));
  sector = (char *) malloc (strlen ("Bottom Right"));
  current_sector = (char *) malloc (strlen ("Bottom Right"));
  strcpy (sector, "Top Left");
  strcpy (current_sector, "Top Left");


  /* Configure Embedded Tcl/Tk and links between C-Tcl/Tk variables. */
  Et_Setup (&argc, argv);
  
  /* Configure the Xlib routines. */
  Xinit (args, argv);

  XSelectInput (Disp, Win, StructureNotifyMask | KeyPressMask | ExposureMask);
  XSelectInput (Disp, FWin, ButtonPressMask | ExposureMask);
  XSelectInput (Disp, RWin,  ButtonPressMask | ExposureMask);
  XSelectInput (Disp, CWin1, ButtonPressMask | ExposureMask);
  XSelectInput (Disp, CWin2, ButtonPressMask | ExposureMask);
  XSelectInput (Disp, CWin3, ButtonPressMask | ExposureMask);
  
  /* Tk responsible for creating and mapping main window.             */
  /* Can use the handle returned from Tk to create the child windows. */
  /* Use Xlib to map the child windows.                               */
  XMapWindow (Disp, FWin);
  XMapWindow (Disp, RWin);
  XMapWindow (Disp, CWin1); 
  XMapWindow (Disp, CWin2); 
  XMapWindow (Disp, CWin3); 
  
  ET( StatusWindow .l -bd 1 -relief ridge \
      -width %d(IMAGEX/2) \
      -height %d((int)((float)(STATUS_HEIGHT-IMAGEY)/14.2)) \
      -bg "light steel blue" \
      -wrap none \
      -font -adobe-helvetica-bold-r-*-*-12-*-*-*-*-*-*-*
      place .l -in . \
      -x %d(4*BORDER+MAXX*IMAGEX-1) \
      -y %d(BORDER+MENU_HEIGHT-1) );
  
  setupPixmaps ();

#ifdef SOUND
#ifdef SOUNDCARD
  setUpAudioSystem ();
#endif
#endif

  /* TCP/IP initialization. */
  do {
    port = (random() % (65535-BASE_PORT)) + BASE_PORT;
  }
  while (Net_Initialize (port) == NET_SERVER_ID_INUSE);

  if (Net_Open (server, SERVER_PORT, &sconn_id) != NET_SUCCESS) {
    fprintf (stderr, "%s: Failed to open connection to %s on port %d.\n", 
	     progname,
	     server, 
	     SERVER_PORT);
    shutdown (1);
  }

  eventHandle (port);
}
